#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False

from libcpp.vector cimport vector
from libc.math cimport sqrt
cimport cython

import numpy as np

import pystan.constants


# autocovariance is a template function, which Cython doesn't yet support
cdef extern from "stan/math/prim/mat/fun/autocovariance.hpp" namespace "stan::math":
    void stan_autocovariance "stan::math::autocovariance<double>"(const vector[double]& y, vector[double]& acov)

cdef extern from "stan/math/prim/mat/fun/sum.hpp" namespace "stan::math":
    double stan_sum "stan::math::sum"(vector[double]& x)

cdef extern from "stan/math/prim/mat/fun/mean.hpp" namespace "stan::math":
    double stan_mean "stan::math::mean"(vector[double]& x)

cdef extern from "stan/math/prim/mat/fun/variance.hpp" namespace "stan::math":
    double stan_variance "stan::math::variance"(vector[double]& x)


cdef void get_kept_samples(dict sim, int k, int n, vector[double]& samples):
    """

    Parameters
    ----------
    k : unsigned int
        Chain index
    n : unsigned int
        Parameter index
    """
    cdef int i
    cdef long[:] warmup2 = np.array(sim['warmup2'])

    slst = sim['samples'][k]['chains']  # chain k, an OrderedDict
    param_names = list(slst.keys())  # e.g., 'beta[1]', 'beta[2]', ...
    cdef double[:] nv = slst[param_names[n]]  # parameter n
    samples.clear()
    for i in range(nv.shape[0] - warmup2[k]):
        samples.push_back(nv[warmup2[k] + i])


cdef double get_chain_mean(dict sim, int k, int n):
    cdef long[:] warmup2 = np.array(sim['warmup2'])
    slst = sim['samples'][k]['chains']  # chain k, an OrderedDict
    param_names = list(slst.keys())  # e.g., 'beta[1]', 'beta[2]', ...
    cdef vector[double] nv = slst[param_names[n]]  # parameter n
    return stan_mean(nv[warmup2[k]:])


cdef vector[double] autocovariance(dict sim, int k, int n):
    """
    Returns the autocovariance for the specified parameter in the
    kept samples of the chain specified.

    Parameters
    ----------
    k : unsigned int
        Chain index
    n : unsigned int
        Parameter index

    Returns
    -------
    acov : vector[double]

    Note
    ----
    PyStan is profligate with memory here in comparison to RStan. A variety
    of copies are made where RStan passes around references. This is done
    mainly for convenience; the Cython code is simpler.
    """
    cdef vector[double] samples, acov
    get_kept_samples(sim, k, n, samples)
    stan_autocovariance(samples, acov)
    return acov


@cython.cdivision(True)
def effective_sample_size(dict sim, int n):
    """
    Return the effective sample size for the specified parameter
    across all kept samples.

    The implementation is close to the effective sample size
    description in BDA3 (p. 286-287).  See more details in Stan
    reference manual section "Effective Sample Size".

    Current implementation takes the minimum number of samples
    across chains as the number of samples per chain.

    Parameters
    ----------
    sim : dict
        Contains samples as well as related information (warmup, number
        of iterations, etc).
    n : int
        Parameter index

    Returns
    -------
    ess : int
    """
    cdef int i, chain
    cdef int m = sim['chains']

    cdef vector[int] ns_save = sim['n_save']
    cdef vector[int] ns_warmup2 = sim['warmup2']
    cdef vector[int] ns_kept = [s - w for s, w in zip(sim['n_save'], sim['warmup2'])]

    cdef int n_samples = min(ns_kept)

    cdef vector[vector[double]] acov
    for chain in range(m):
        acov.push_back(autocovariance(sim, chain, n))

    cdef vector[double] chain_mean
    cdef vector[double] chain_var
    # double rather than int to deal with Cython quirk, see issue #186
    cdef double n_kept_samples
    for chain in range(m):
        n_kept_samples = ns_kept[chain]
        if n_kept_samples == 1:
            # fix crash for mingw on Windows
            return np.nan
        chain_mean.push_back(get_chain_mean(sim, chain, n))
        chain_var.push_back(acov[chain][0] * n_kept_samples / (n_kept_samples-1))

    cdef double mean_var = stan_mean(chain_var)
    cdef double var_plus = mean_var * (n_samples-1) / n_samples

    if m > 1:
        var_plus = var_plus + stan_variance(chain_mean)

    cdef vector[double] rho_hat_t
    for _ in range(n_samples):
        rho_hat_t.push_back(0)
    cdef vector[double] acov_t
    acov_t.clear()
    for chain in range(m):
        acov_t.push_back(acov[chain][1])
    cdef double rho_hat_even = 1
    rho_hat_t[0] = rho_hat_even
    cdef double rho_hat_odd = 1 - (mean_var - stan_mean(acov_t)) / var_plus
    rho_hat_t[1] = rho_hat_odd
    # Geyer's initial positive sequence
    cdef int max_t = 1
    cdef int t = 1
    while t < (n_samples - 2) and (rho_hat_even + rho_hat_odd) >= 0:
        acov_t.clear()
        for chain in range(m):
            acov_t.push_back(acov[chain][t + 1])
        rho_hat_even = 1 - (mean_var - stan_mean(acov_t)) / var_plus
        acov_t.clear()
        for chain in range(m):
            acov_t.push_back(acov[chain][t + 2])
        rho_hat_odd = 1 - (mean_var - stan_mean(acov_t)) / var_plus
        if (rho_hat_even + rho_hat_odd) >= 0:
            rho_hat_t[t + 1] = rho_hat_even
            rho_hat_t[t + 2] = rho_hat_odd
        max_t = t + 2
        t += 2
    # Geyer's initial monotone sequence
    t = 3
    while t <= max_t - 2:
        if rho_hat_t[t + 1] + rho_hat_t[t + 2] > rho_hat_t[t - 1] + rho_hat_t[t]:
            rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2
            rho_hat_t[t + 2] = rho_hat_t[t + 1]
        t += 2
    cdef double ess = m * n_samples
    ess = ess / (-1 + 2 * stan_sum(rho_hat_t))
    return ess


@cython.cdivision(True)
def split_potential_scale_reduction(dict sim, int n):
    """
    Return the split potential scale reduction (split R hat) for the
    specified parameter.

    Current implementation takes the minimum number of samples
    across chains as the number of samples per chain.

    Parameters
    ----------
    n : unsigned int
        Parameter index

    Returns
    -------
    rhat : float
        Split R hat

    """
    cdef int i, chain
    cdef double srhat

    cdef int n_chains = sim['chains']

    cdef vector[int] ns_save = sim['n_save']
    cdef vector[int] ns_warmup2 = sim['warmup2']
    cdef vector[int] ns_kept = [s - w for s, w in zip(sim['n_save'], sim['warmup2'])]

    cdef int n_samples = min(ns_kept)

    if n_samples % 2 == 1:
        n_samples = n_samples - 1

    cdef vector[double] split_chain_mean, split_chain_var
    cdef vector[double] samples, split_chain
    for chain in range(n_chains):
        samples.clear()
        get_kept_samples(sim, chain, n, samples)
        split_chain.clear()
        for i in range(n_samples / 2):
            split_chain.push_back(samples[i])
        split_chain_mean.push_back(stan_mean(split_chain))
        split_chain_var.push_back(stan_variance(split_chain))

        split_chain.clear()
        for i in range(n_samples / 2, n_samples):
            split_chain.push_back(samples[i])
        split_chain_mean.push_back(stan_mean(split_chain))
        split_chain_var.push_back(stan_variance(split_chain))

    cdef double var_between = n_samples / 2 * stan_variance(split_chain_mean)
    cdef double var_within = stan_mean(split_chain_var)

    srhat = sqrt((var_between / var_within + n_samples / 2 - 1) / (n_samples / 2))
    return srhat


def _test_autocovariance(dict sim, int k, int n):
    '''Test point for autocovariance function'''
    return autocovariance(sim, k, n)


def _test_stan_functions():
    y = np.arange(10)
    cdef vector[double] acov
    stan_autocovariance(y, acov)
    assert sum(acov) == -40.0, sum(acov)
    assert stan_sum(y) == sum(y)
    assert stan_mean(y) == np.mean(y)
    assert stan_variance(y) == np.var(y, ddof=1)



# distutils: language = c++
# cython: language_level=2
#-----------------------------------------------------------------------------
# Copyright (c) 2013, Allen B. Riddell
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# THIS IS A TEMPLATE, not a proper Cython .pyx file
#
# A template variable $model_cppname needs to be substituted before
# compilation.
#
# $model_cppname.hpp should be in the same directory as this
# file (after substitutions have been made).
#
#-----------------------------------------------------------------------------

# cython imports
from libcpp cimport bool
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.string cimport string
from libcpp.vector cimport vector

from cython.operator cimport dereference as deref, preincrement as inc

cimport numpy as np

from pystan.io cimport py_var_context, var_context
from pystan.stan_fit cimport stan_fit, StanArgs, StanHolder, get_all_flatnames

# Initialize numpy for use from C. When using numpy from C or Cython this must always be done.
np.import_array()

# python imports
from collections import OrderedDict
import logging
import warnings

import numpy as np

import pystan.misc
import pystan.plots
from pystan._compat import PY2, string_types
from pystan.constants import (sampling_algo_t, optim_algo_t, variational_algo_t,
                              sampling_metric_t, stan_args_method_t)

logger = logging.getLogger('pystan')

cdef extern from "boost/random/additive_combine.hpp" namespace "boost::random":
    cdef cppclass additive_combine_engine[T, U]:
        pass
    ctypedef additive_combine_engine ecuyer1988

cdef extern from "$model_cppname.hpp" namespace "${model_cppname}_namespace":
    cdef cppclass $model_cppname:
        $model_cppname(var_context& context) except +

# NOTE: Methods that aren't intended for public use are prefixed by '_'. For
# example, _update_param_oi probably shouldn't be called unless you know
# something about the state of the C++ class instance wrapped by the class.

ctypedef map[string, pair[vector[double], vector[size_t]]] vars_r_t
ctypedef map[string, pair[vector[int], vector[size_t]]] vars_i_t


cdef class PyStanHolder:
    """Allow access to a StanHolder instance from Python

    A PyStanHolder instance wraps a StanHolder instance. When the PyStanHolder
    instance is deleted, the StanHolder instance will be as well.

    There are slight differences between the StanHolder and PyStanHolder. For
    example, chains is an OrderedDict in the PyStanHolder where a StanHolder
    tracks the same information in the fields ``chains`` and ``chain_names``.
    The same holds for ``sampler_params``.
    """
    cdef public int num_failed
    cdef public bool test_grad
    cdef public list inits
    cdef public list par
    cdef public double value
    cdef public chains
    cdef public dict args
    cdef public mean_pars
    cdef public list mean_par_names
    cdef public double mean_lp__
    cdef public adaptation_info
    cdef public sampler_params
    cdef public list sampler_param_names
    cdef StanHolder * holderptr

    # for backward compatibility allow holder[attr]
    def __getitem__(self, key):
        return getattr(self, key)

    def __dealloc__(self):
        del self.holderptr

    # the following three methods give Cython classes instructions for pickling
    def __getstate__(self):
        attr_names = ('num_failed test_grad inits par value chains args mean_pars mean_lp__ '
                      'adaptation_info sampler_params sampler_param_names').split()
        state = dict((k, getattr(self, k)) for k in attr_names)
        return state

    def __setstate__(self, state):
        for k in state:
            setattr(self, k, state[k])

    def __reduce__(self):
        return (PyStanHolder, tuple(), self.__getstate__(), None, None)


cdef PyStanHolder _pystanholder_from_stanholder(StanHolder* holder):
    cdef int num_iter
    cdef double* data_ptr
    cdef np.npy_intp dims[1]

    h = PyStanHolder()
    h.holderptr = holder
    h.num_failed = holder.num_failed
    h.test_grad = holder.test_grad
    h.inits = holder.inits
    h.par = holder.par
    h.value = holder.value

    chains = []
    cdef vector[vector[double] ].iterator it = holder.chains.begin()
    while it != holder.chains.end():
        num_iter = deref(it).size()
        dims[0] = <np.npy_intp> num_iter
        data_ptr = &(deref(it).front())
        ch = np.PyArray_SimpleNewFromData(1, dims, np.NPY_DOUBLE, data_ptr)
        chains.append(ch)
        inc(it)
    chain_names = [n.decode('utf-8') for n in holder.chain_names]
    h.chains = OrderedDict(zip(chain_names, chains))

    # NOTE: when _pystanholder_from_stanholder is called we also have a pointer
    # to holder.args available so we will use it directly from there. Strictly
    # speaking it should be done here, but Cython kept throwing errors
    # FIXME: figure out origins of difficulties
    # r['args'] = _dict_from_stanargs(holder.args)
    h.mean_pars = holder.mean_pars
    h.mean_par_names = [n.decode('utf-8') for n in holder.mean_par_names]
    h.mean_lp__ = holder.mean_lp__
    h.adaptation_info = holder.adaptation_info.decode('utf-8')
    h.sampler_params = holder.sampler_params
    h.sampler_param_names = [n.decode('utf-8') for n in holder.sampler_param_names]
    return h


cdef dict _dict_from_stanargs(StanArgs* args):
    d = dict()
    ctrl_d = dict()
    d['random_seed'] = str(args.random_seed)
    d['chain_id'] = args.chain_id
    d['init'] = args.init
    # FIXME: reconstructing d['init_list'] from args.init_vars_r and
    # args.init_vars_i requires additional work. The initial values for each
    # chain are accessible with the method get_inits()
    d['init_radius'] = args.init_radius
    d['enable_random_init'] = args.enable_random_init
    d['append_samples'] = args.append_samples
    if args.sample_file_flag:
        d['sample_file'] = args.sample_file
    if args.diagnostic_file_flag:
        d['diagnostic_file'] = args.diagnostic_file

    method = stan_args_method_t(args.method)
    if method == stan_args_method_t.SAMPLING:
        d["method"] = method.name
        d["iter"] = args.ctrl.sampling.iter
        d["warmup"] = args.ctrl.sampling.warmup
        d["thin"] = args.ctrl.sampling.thin
        d["refresh"] = args.ctrl.sampling.refresh
        d["test_grad"] = False
        ctrl_d["adapt_engaged"] = args.ctrl.sampling.adapt_engaged
        ctrl_d["adapt_gamma"] = args.ctrl.sampling.adapt_gamma
        ctrl_d["adapt_delta"] = args.ctrl.sampling.adapt_delta
        ctrl_d["adapt_kappa"] = args.ctrl.sampling.adapt_kappa
        ctrl_d["adapt_init_buffer"] = args.ctrl.sampling.adapt_init_buffer
        ctrl_d["adapt_term_buffer"] = args.ctrl.sampling.adapt_term_buffer
        ctrl_d["adapt_window"] = args.ctrl.sampling.adapt_window
        ctrl_d["adapt_t0"] = args.ctrl.sampling.adapt_t0
        ctrl_d["stepsize"] = args.ctrl.sampling.stepsize
        ctrl_d["stepsize_jitter"] = args.ctrl.sampling.stepsize_jitter
        d["sampler_t"] = algorithm = sampling_algo_t(args.ctrl.sampling.algorithm).name
        if algorithm == sampling_algo_t.NUTS:
            ctrl_d["max_treedepth"] = args.ctrl.sampling.max_treedepth
        elif algorithm == sampling_algo_t.HMC:
            ctrl_d["int_time"] = args.ctrl.sampling.int_time
        elif algorithm == sampling_algo_t.Metropolis:
            pass
        else:
            # included here to mirror rstan code
            pass
        if algorithm != sampling_algo_t.Metropolis:
            metric = sampling_metric_t(args.ctrl.sampling.metric).name
            if metric == sampling_metric_t.UNIT_E:
                ctrl_d["metric"] = "unit_e"
                d["sampler_t"] = d["sampler_t"] + "(unit_e)"
            elif metric == sampling_metric_t.DIAG_E:
                ctrl_d["metric"] = "diag_e"
                d["sampler_t"] = d["sampler_t"] + "(diag_e)"
            elif metric == sampling_metric_t.DENSE_E:
                ctrl_d["metric"] = "dense_e"
                d["sampler_t"] = d["sampler_t"] + "(dense_e)"
        d["control"] = ctrl_d
    elif method == stan_args_method_t.VARIATIONAL:
        d["method"] = method.name
        d["iter"] = args.ctrl.variational.iter
        d["grad_samples"] = args.ctrl.variational.grad_samples
        d["elbo_samples"] = args.ctrl.variational.elbo_samples
        d["eval_elbo"] = args.ctrl.variational.eval_elbo
        d["output_samples"] = args.ctrl.variational.output_samples
        d["eta"] = args.ctrl.variational.eta
        d["adapt_engaged"] = args.ctrl.variational.adapt_engaged
        d["adapt_iter"] = args.ctrl.variational.adapt_iter
        d["tol_rel_obj"] = args.ctrl.variational.tol_rel_obj
        algorithm = variational_algo_t(args.ctrl.variational.algorithm)
        d['algorithm'] = algorithm.name
    elif method == stan_args_method_t.OPTIM:
        d["method"] = method.name
        d["iter"] = args.ctrl.optim.iter
        d["refresh"] = args.ctrl.optim.refresh
        d["save_iterations"] = args.ctrl.optim.save_iterations
        algorithm = optim_algo_t(args.ctrl.optim.algorithm)
        d["algorithm"] = algorithm.name
        if algorithm == optim_algo_t.Newton:
            pass
        elif algorithm == optim_algo_t.LBFGS:
            d["init_alpha"] = args.ctrl.optim.init_alpha
            d["tol_param"] = args.ctrl.optim.tol_param
            d["tol_obj"] = args.ctrl.optim.tol_obj
            d["tol_grad"] = args.ctrl.optim.tol_grad
            d["tol_rel_obj"] = args.ctrl.optim.tol_obj
            d["tol_rel_grad"] = args.ctrl.optim.tol_grad
            d["tol_history_size"] = args.ctrl.optim.tol_grad
        elif algorithm == optim_algo_t.BFGS:
            d["init_alpha"] = args.ctrl.optim.init_alpha
            d["tol_param"] = args.ctrl.optim.tol_param
            d["tol_obj"] = args.ctrl.optim.tol_obj
            d["tol_grad"] = args.ctrl.optim.tol_grad
            d["tol_rel_obj"] = args.ctrl.optim.tol_obj
            d["tol_rel_grad"] = args.ctrl.optim.tol_grad
    elif method == stan_args_method_t.TEST_GRADIENT:
        d["method"] = "test_grad"
        d["test_grad"] = True
        ctrl_d["epsilon"] = args.ctrl.test_grad.epsilon
        ctrl_d["error"] = args.ctrl.test_grad.error
        d["control"] = ctrl_d
    return d


cdef void _set_stanargs_from_dict(StanArgs* p, dict args):
    """Insert values in dictionary `args` into `p`"""
    # _call_sampler requires a specially crafted dictionary of arguments
    # intended for the c++ function sampler_command(...) in stan_fit.hpp
    # If the dictionary doesn't contain the correct keys (arguments),
    # the function will raise a KeyError exception (as it should!).
    cdef vars_r_t init_vars_r
    cdef vars_i_t init_vars_i
    p.random_seed = <unsigned int> args.get('random_seed', 0)
    p.chain_id = <unsigned int> args['chain_id']
    p.init = args['init']
    if args['init'] == b'user':
        init_r, init_i = pystan.misc._split_data(args['init_list'])
        init_vars_r = _dict_to_vars_r(init_r)
        init_vars_i = _dict_to_vars_i(init_i)
        p.init_vars_r = init_vars_r
        p.init_vars_i = init_vars_i
    p.init_radius = args['init_radius']
    p.sample_file = args['sample_file']
    p.append_samples = args['append_samples']
    p.sample_file_flag = args['sample_file_flag']
    p.method = args['method'].value
    p.diagnostic_file = args['diagnostic_file']
    p.diagnostic_file_flag = args['diagnostic_file_flag']
    p.metric_file = args['metric_file']
    p.metric_file_flag = args['metric_file_flag']
    if args['method'] == stan_args_method_t.SAMPLING:
        p.ctrl.sampling.iter = args['ctrl']['sampling']['iter']
        p.ctrl.sampling.refresh = args['ctrl']['sampling']['refresh']
        p.ctrl.sampling.algorithm = args['ctrl']['sampling']['algorithm'].value
        p.ctrl.sampling.warmup = args['ctrl']['sampling']['warmup']
        p.ctrl.sampling.thin = args['ctrl']['sampling']['thin']
        p.ctrl.sampling.save_warmup = args['ctrl']['sampling']['save_warmup']
        p.ctrl.sampling.iter_save = args['ctrl']['sampling']['iter_save']
        p.ctrl.sampling.iter_save_wo_warmup = args['ctrl']['sampling']['iter_save_wo_warmup']
        p.ctrl.sampling.adapt_engaged = args['ctrl']['sampling']['adapt_engaged']
        p.ctrl.sampling.adapt_gamma = args['ctrl']['sampling']['adapt_gamma']
        p.ctrl.sampling.adapt_delta = args['ctrl']['sampling']['adapt_delta']
        p.ctrl.sampling.adapt_kappa = args['ctrl']['sampling']['adapt_kappa']
        p.ctrl.sampling.adapt_init_buffer = args['ctrl']['sampling']['adapt_init_buffer']
        p.ctrl.sampling.adapt_term_buffer = args['ctrl']['sampling']['adapt_term_buffer']
        p.ctrl.sampling.adapt_window = args['ctrl']['sampling']['adapt_window']
        p.ctrl.sampling.adapt_t0 = args['ctrl']['sampling']['adapt_t0']
        p.ctrl.sampling.metric = args['ctrl']['sampling']['metric'].value
        p.ctrl.sampling.stepsize = args['ctrl']['sampling']['stepsize']
        p.ctrl.sampling.stepsize_jitter = args['ctrl']['sampling']['stepsize_jitter']
        if args['ctrl']['sampling']['algorithm'] == sampling_algo_t.NUTS:
            p.ctrl.sampling.max_treedepth = args['ctrl']['sampling']['max_treedepth']
        elif args['ctrl']['sampling']['algorithm'] == sampling_algo_t.HMC:
            p.ctrl.sampling.int_time = args['ctrl']['sampling']['int_time']
    elif args['method'] == stan_args_method_t.OPTIM:
        p.ctrl.optim.iter = args['ctrl']['optim']['iter']
        p.ctrl.optim.refresh = args['ctrl']['optim']['refresh']
        p.ctrl.optim.algorithm = args['ctrl']['optim']['algorithm'].value
        p.ctrl.optim.save_iterations = args['ctrl']['optim']['save_iterations']
        p.ctrl.optim.init_alpha = args['ctrl']['optim']['init_alpha']
        p.ctrl.optim.tol_obj = args['ctrl']['optim']['tol_obj']
        p.ctrl.optim.tol_grad = args['ctrl']['optim']['tol_grad']
        p.ctrl.optim.tol_param = args['ctrl']['optim']['tol_param']
        p.ctrl.optim.tol_rel_obj = args['ctrl']['optim']['tol_rel_obj']
        p.ctrl.optim.tol_rel_grad = args['ctrl']['optim']['tol_rel_grad']
        p.ctrl.optim.history_size = args['ctrl']['optim']['history_size']
    elif args['method'] == stan_args_method_t.TEST_GRADIENT:
        p.ctrl.test_grad.epsilon = args['ctrl']['test_grad']['epsilon']
        p.ctrl.test_grad.error = args['ctrl']['test_grad']['error']
    elif args['method'] == stan_args_method_t.VARIATIONAL:
        p.ctrl.variational.algorithm = args['ctrl']['variational']['algorithm'].value
        p.ctrl.variational.iter = args['ctrl']['variational']['iter']
        p.ctrl.variational.grad_samples = args['ctrl']['variational']['grad_samples']
        p.ctrl.variational.elbo_samples = args['ctrl']['variational']['elbo_samples']
        p.ctrl.variational.eval_elbo = args['ctrl']['variational']['eval_elbo']
        p.ctrl.variational.output_samples = args['ctrl']['variational']['output_samples']
        p.ctrl.variational.eta = args['ctrl']['variational']['eta']
        p.ctrl.variational.adapt_engaged = args['ctrl']['variational']['adapt_engaged']
        p.ctrl.variational.tol_rel_obj = args['ctrl']['variational']['tol_rel_obj']
        p.ctrl.variational.adapt_iter = args['ctrl']['variational']['adapt_iter']

cdef vars_r_t _dict_to_vars_r(data_r):
    """Converts a dict or OrderedDict to a C++ map of string, double pairs"""
    cdef vars_r_t vars_r

    # The dimension for a single value is an empty vector. A list of
    # values is indicated by an entry with the number of values.
    # The dimensions of an array are indicated as one would expect.
    #
    # note, array.flat yields values in C-contiguous style, with the
    # last index varying the fastest. So the transpose is taken
    # so that the ordering matches that used by stan.
    for key in data_r:
        assert isinstance(key, bytes), "Variable name must be bytes."
        val = (data_r[key].T.flat, data_r[key].shape)
        vars_r[key] = val
    return vars_r


cdef vars_i_t _dict_to_vars_i(data_i):
    """Converts a dict or OrdereDict to a C++ map of string, int pairs"""
    cdef vars_i_t vars_i

    # The dimension for a single value is an empty vector. A list of
    # values is indicated by an entry with the number of values.
    # The dimensions of an array are indicated as one would expect.
    #
    # note, array.flat yields values in C-contiguous style, with the
    # last index varying the fastest. So the transpose is taken
    # so that the ordering matches that used by stan.
    for key in data_i:
        assert isinstance(key, bytes), "Variable name must be bytes."
        val = (data_i[key].T.flat, data_i[key].shape)
        vars_i[key] = val
    return vars_i

def _call_sampler_star(data_args):
    return _call_sampler(*data_args)

def _call_sampler(data, args, pars_oi=None):
    """Wrapper for call_sampler in stan_fit

    This function is self-contained and suitable for parallel invocation.

    """
    data_r, data_i = pystan.misc._split_data(data)
    cdef StanHolder *holderptr = new StanHolder()
    cdef StanArgs *argsptr = new StanArgs()
    if not holderptr:
        raise MemoryError("Couldn't allocate space for StanHolder.")
    if not argsptr:
        raise MemoryError("Couldn't allocate space for StanArgs.")
    chain_id = args['chain_id']
    for handler in logger.handlers:
        handler.flush()
    _set_stanargs_from_dict(argsptr, args)

    cdef stan_fit[$model_cppname, ecuyer1988] *fitptr
    cdef vars_r_t vars_r = _dict_to_vars_r(data_r)
    cdef vars_i_t vars_i = _dict_to_vars_i(data_i)
    fitptr = new stan_fit[$model_cppname, ecuyer1988](vars_r, vars_i, argsptr.random_seed)
    if not fitptr:
        raise MemoryError("Couldn't allocate space for stan_fit.")
    # Implementation note: there is an extra stan_fit instance associated
    # with the model (which enables access to some methods). This is a
    # horrible, confusing idea which will hopefully be fixed in Stan 3.
    if pars_oi is not None:
        pars_oi_bytes = [n.encode('ascii') for n in pars_oi]
        if len(pars_oi_bytes) != fitptr.param_names_oi().size():
            fitptr.update_param_oi(pars_oi_bytes)
    ret = fitptr.call_sampler(deref(argsptr), deref(holderptr))
    holder = _pystanholder_from_stanholder(holderptr)
    # FIXME: rather than fetching the args from the holderptr, we just use
    # the argsptr we passed directly. This is a hack to solve a problem
    # that holder.args gets dropped somewhere in C++.
    holder.args = _dict_from_stanargs(argsptr)
    del argsptr
    del fitptr
    return ret, holder

def _split_pars_locs(fnames, pars):
    """Split flatnames to par and location"""
    par_keys = OrderedDict((par, []) for par in pars)
    for key in fnames:
        par_tail = key.split("[")
        par = par_tail[0]
        loc = [Ellipsis]
        for tail in par_tail[1:]:
            loc = []
            for i in tail[:-1].split(","):
                loc.append(int(i)-1)
        par_keys[par].append((key, loc))
    return par_keys

cdef class StanFit4Model:
    """Holder for results obtained from running a Stan model with data

    Attributes
    ----------
    sim : dict
        Holder for runs. Stores samples in sim['samples']
    data : dict
        Data used to fit model.

    Note
    ----
    The only unexpected difference between PyStan and RStan is this: where RStan
    stores samples for a parameter directly in, say, fit@sim$samples[[1]]$theta,
    in PyStan they are in fit.sim['samples'][0]['chains']['theta'].

    The difference is due to Python lacking a dictionary structure that can also
    have attributes.

    """

    cdef stan_fit[$model_cppname, ecuyer1988] *thisptr

    # attributes populated by methods of StanModel
    cdef public data  # dict or OrderedDict
    cdef public random_seed
    cdef public dict sim
    cdef public model_name
    cdef public model_pars
    cdef public par_dims
    cdef public mode
    cdef public inits
    cdef public stan_args
    cdef public stanmodel
    cdef public date
    cdef public _repr_pars
    cdef public _repr_num

    def __cinit__(self, *args):
        # __cinit__ must be callable with no arguments for unpickling
        cdef vars_r_t vars_r
        cdef vars_i_t vars_i
        if len(args) == 2:
            data, random_seed = args
            data_r, data_i = pystan.misc._split_data(data)
            # NB: dictionary keys must be byte strings
            vars_r = _dict_to_vars_r(data_r)
            vars_i = _dict_to_vars_i(data_i)
            # TODO: the random seed needs to be known by StanFit4Model
            self.thisptr = new stan_fit[$model_cppname, ecuyer1988](vars_r, vars_i, <unsigned int> random_seed)
            if not self.thisptr:
                raise MemoryError("Couldn't allocate space for stan_fit.")

    def __init__(self, data, random_seed):
        self.data = data
        self.random_seed = random_seed
        self._set_repr_pars(None)
        self._set_repr_num (100)

    def __dealloc__(self):
        del self.thisptr

    # the following three methods give Cython classes instructions for pickling
    def __getstate__(self):
        attr_names = ('data sim model_name model_pars par_dims mode inits stan_args '
                      'stanmodel date').split()
        state = dict((k, getattr(self, k)) for k in attr_names)
        return state

    def __setstate__(self, state):
        for k in state:
            setattr(self, k, state[k])

    def __reduce__(self):
        msg = ("Pickling fit objects is an experimental feature!\n"
               "The relevant StanModel instance must be pickled along with this fit object.\n"
               "When unpickling the StanModel must be unpickled first.")
        warnings.warn(msg)
        return (StanFit4Model, (self.data, self.random_seed), self.__getstate__(), None, None)

    # public methods

    def plot(self, pars=None, dtypes=None):
        """Visualize samples from posterior distributions

        Parameters
        ---------
        pars : {str, sequence of str}
            parameter name(s); by default use all parameters of interest
        dtypes : dict
            datatype of parameter(s).
            If nothing is passed, np.float will be used for all parameters.
            If np.int is specified, the histogram will be visualized, not but
            kde.

        Note
        ----
        This is currently an alias for the `traceplot` method.
        """
        if pars is None:
            pars = [par for par in self.sim['pars_oi'] if par != 'lp__']
        elif isinstance(pars, string_types):
            pars = [pars]
        pars = pystan.misc._remove_empty_pars(pars, self.sim['pars_oi'], self.sim['dims_oi'])
        return pystan.plots.traceplot(self, pars, dtypes)

    def traceplot(self, pars=None, dtypes=None):
        """Visualize samples from posterior distributions

        Parameters
        ---------
        pars : {str, sequence of str}, optional
            parameter name(s); by default use all parameters of interest
        dtypes : dict
            datatype of parameter(s).
            If nothing is passed, np.float will be used for all parameters.
            If np.int is specified, the histogram will be visualized, not but
            kde.
        """
        # FIXME: for now plot and traceplot do the same thing
        return self.plot(pars, dtypes=dtypes)

    def extract(self, pars=None, permuted=True, inc_warmup=False, dtypes=None):
        """Extract samples in different forms for different parameters.

        Parameters
        ----------
        pars : {str, sequence of str}
           parameter (or quantile) name(s).
        permuted : bool
           If True, returned samples are permuted. All chains are
           merged and warmup samples are discarded.
        inc_warmup : bool
           If True, warmup samples are kept; otherwise they are
           discarded. If `permuted` is True, `inc_warmup` is ignored.
        dtypes : dict
            datatype of parameter(s).
            If nothing is passed, np.float will be used for all parameters.

        Returns
        -------
        samples : dict or array
        If `permuted` is True, return dictionary with samples for each
        parameter (or other quantity) named in `pars`.

        If `permuted` is False and `pars` is None, an array is returned. The first dimension of
        the array is for the iterations; the second for the number of chains;
        the third for the parameters. Vectors and arrays are expanded to one
        parameter (a scalar) per cell, with names indicating the third dimension.
        Parameters are listed in the same order as `model_pars` and `flatnames`.

        If `permuted` is False and `pars` is not None, return dictionary with samples for each
        parameter (or other quantity) named in `pars`. The first dimension of
        the sample array is for the iterations; the second for the number of chains;
        the rest for the parameters. Parameters are listed in the same order as `pars`.

        """
        self._verify_has_samples()
        if inc_warmup is True and permuted is True:
            logger.warning("`inc_warmup` ignored when `permuted` is True.")
            inc_warmup = False
        if dtypes is not None and permuted is False and pars is None:
            logger.warning("`dtypes` ignored when `permuted` is False and `pars` is None")

        pars_original = pars
        if pars is None:
            pars = self.sim['pars_oi']
        elif isinstance(pars, string_types):
            pars = [pars]
        pars = pystan.misc._remove_empty_pars(pars, self.sim['pars_oi'], self.sim['dims_oi'])
        if dtypes is None:
            dtypes = {}

        allpars = self.sim['pars_oi'] + self.sim['fnames_oi']
        pystan.misc._check_pars(allpars, pars)

        n_kept = [s if inc_warmup else s-w for s, w in zip(self.sim['n_save'], self.sim['warmup2'])]
        chains = len(self.sim['samples'])

        # return array (n, chains, flat_pars)
        if (not permuted) and (pars_original is None):
            n = n_kept[0]
            arr_shape = [n, chains, len(self.sim['fnames_oi'])]
            arr = np.empty(arr_shape, order='F')
            for chain, (pyholder, n) in enumerate(zip(self.sim['samples'], n_kept)):
                for i, item in enumerate(pyholder.chains.values()):
                    arr[:, chain, i] = item[-n:]
            return arr


        par_keys = _split_pars_locs(self.sim['fnames_oi'], self.sim['pars_oi'])

        shapes = dict(zip(self.sim['pars_oi'], self.sim['dims_oi']))

        extracted = OrderedDict()
        for par in pars:
            if par in extracted:
                continue
            keys_locs = par_keys.get(par, [(par, [Ellipsis])])
            shape = shapes.get(par, [])
            dtype = dtypes.get(par)

            if permuted:
                arr_shape = [sum(n_kept)] + shape
                arr = np.empty(arr_shape, dtype=dtype, order='F')
                for chain, (pyholder, permutation, n) in enumerate(zip(self.sim['samples'], self.sim['permutation'], n_kept)):
                    n_processed = sum(n_kept[:chain])
                    axes = [slice(n_processed, n_processed+n)]
                    for key, loc in keys_locs:
                        arr_slice = tuple(axes + loc)
                        arr[arr_slice] = pyholder.chains[key][-n:][permutation]
                extracted[par] = arr

            else:
                n = n_kept[0]
                arr_shape = [n, chains] + shape
                arr = np.empty(arr_shape, dtype=dtype, order='F')
                for chain, (pyholder, n) in enumerate(zip(self.sim['samples'], n_kept)):
                    axes = [slice(None), chain]
                    for key, loc in keys_locs:
                        arr_slice = tuple(axes + loc)
                        arr[arr_slice] = pyholder.chains[key][-n:]
                extracted[par] = arr

        return extracted

    def __unicode__(self):
        # for Python 2.x
        return pystan.misc.stansummary(self)

    def __str__(self):
        s = pystan.misc.stansummary(self)
        return s.encode('utf-8') if PY2 else s

    def _get_repr_num(self):
      return self._repr_num

    def _get_repr_pars(self):
        update = False
        if self._repr_pars is not None:
            update = len(self._repr_pars)-1 != self._repr_num
        if self._repr_pars is not None and not update:
            return self._repr_pars
        elif len(self.sim["fnames_oi"]) > self._repr_num:
            logger.warning("Truncated summary with the 'fit.__repr__' method. For the full summary use 'print(fit)'")
            self._set_repr_pars(self.sim["fnames_oi"][:self._repr_num-1] + ['lp__'])
            return self._repr_pars

    def _set_repr_num(self, n):
        self._repr_num = n

    def _set_repr_pars(self, pars):
        self._repr_pars = pars

    def __repr__(self):
        pars = self._get_repr_pars()
        if pars is not None and len(self.sim["fnames_oi"]) > len(pars):
            s = "\nWarning: Shown data is truncated to {} parameters".format(len(pars))
            s += "\nFor the full summary use 'print(fit)'\n\n"
            s += pystan.misc.stansummary(self, pars=pars)
        else:
            s = pystan.misc.stansummary(self)
        return s.encode('utf-8') if PY2 else s

    def __getitem__(self, key):
        extr = self.extract(pars=(key,))
        return extr[key]

    def stansummary(self, pars=None, probs=(0.025, 0.25, 0.5, 0.75, 0.975), digits_summary=2):
        """
        Summary statistic table.

        Parameters
        ----------
        pars : str or sequence of str, optional
            Parameter names. By default use all parameters
        probs : sequence of float, optional
            Quantiles. By default, (0.025, 0.25, 0.5, 0.75, 0.975)
        digits_summary : int, optional
            Number of significant digits. By default, 2
        Returns
        -------
        summary : string
            Table includes mean, se_mean, sd, probs_0, ..., probs_n, n_eff and Rhat.

        Examples
        --------
        >>> model_code = 'parameters {real y;} model {y ~ normal(0,1);}'
        >>> m = StanModel(model_code=model_code, model_name="example_model")
        >>> fit = m.sampling()
        >>> print(fit.stansummary())
        Inference for Stan model: example_model.
        4 chains, each with iter=2000; warmup=1000; thin=1;
        post-warmup draws per chain=1000, total post-warmup draws=4000.

               mean se_mean     sd   2.5%    25%    50%    75%  97.5%  n_eff   Rhat
        y      0.01    0.03    1.0  -2.01  -0.68   0.02   0.72   1.97   1330    1.0
        lp__   -0.5    0.02   0.68  -2.44  -0.66  -0.24  -0.05-5.5e-4   1555    1.0

        Samples were drawn using NUTS at Thu Aug 17 00:52:25 2017.
        For each parameter, n_eff is a crude measure of effective sample size,
        and Rhat is the potential scale reduction factor on split chains (at
        convergence, Rhat=1).
        """
        return pystan.misc.stansummary(fit=self, pars=pars, probs=probs, digits_summary=digits_summary)

    def summary(self, pars=None, probs=None):
        """Summarize samples (compute mean, SD, quantiles) in all chains.
        REF: stanfit-class.R summary method
        Parameters
        ----------
        pars : str or sequence of str, optional
            Parameter names. By default use all parameters
        probs : sequence of float, optional
            Quantiles. By default, (0.025, 0.25, 0.5, 0.75, 0.975)
        Returns
        -------
        summaries : OrderedDict of array
            Array indexed by 'summary' has dimensions (num_params, num_statistics).
            Parameters are unraveled in *row-major order*. Statistics include: mean,
            se_mean, sd, probs_0, ..., probs_n, n_eff, and Rhat. Array indexed by
            'c_summary' breaks down the statistics by chain and has dimensions
            (num_params, num_statistics_c_summary, num_chains). Statistics for
            `c_summary` are the same as for `summary` with the exception that
            se_mean, n_eff, and Rhat are absent. Row names and column names are
            also included in the OrderedDict.
        """
        return pystan.misc._summary(self, pars, probs)

    def log_prob(self, upar, adjust_transform=True, gradient=False):
        """
        Expose the log_prob of the model to stan_fit so user can call
        this function.

        Parameters
        ----------
        upar : array
            The real parameters on the unconstrained space.
        adjust_transform : bool
            Whether we add the term due to the transform from constrained
            space to unconstrained space implicitly done in Stan.

        Note
        ----
        In Stan, the parameters need be defined with their supports. For
        example, for a variance parameter, we must define it on the positive
        real line. But inside Stan's sampler, all parameters defined on the
        constrained space are transformed to unconstrained space, so the log
        density function need be adjusted (i.e., adding the log of the absolute
        value of the Jacobian determinant).  With the transformation, Stan's
        samplers work on the unconstrained space and once a new iteration is
        drawn, Stan transforms the parameters back to their supports. All the
        transformation are done inside Stan without interference from the users.
        However, when using the log density function for a model exposed to
        Python, we need to be careful.  For example, if we are interested in
        finding the mode of parameters on the constrained space, we then do not
        need the adjustment.  For this reason, there is an argument named
        `adjust_transform` for functions `log_prob` and `grad_log_prob`.

        """
        # gradient is ignored for now. Call grad_log_prob to get the gradient.
        cdef vector[double] par_r = np.asarray(upar).flat
        return self.thisptr.log_prob(par_r, adjust_transform, gradient)

    def grad_log_prob(self, upar, adjust_transform=True):
        """
        Expose the grad_log_prob of the model to stan_fit so user
        can call this function.

        Parameters
        ----------
        upar : array
            The real parameters on the unconstrained space.
        adjust_transform : bool
            Whether we add the term due to the transform from constrained
            space to unconstrained space implicitly done in Stan.
        """
        cdef vector[double] par_r, grad
        par_r = np.asarray(upar).flat
        grad = self.thisptr.grad_log_prob(par_r, adjust_transform)
        return np.asarray(grad)

    def get_adaptation_info(self):
        """Obtain adaptation information for sampler, which now only NUTS2 has.

        The results are returned as a list, each element of which is a character
        string for a chain."""
        self._verify_has_samples()
        lai =  [ch['adaptation_info'] for ch in self.sim['samples']]
        return lai

    def get_logposterior(self, inc_warmup=True):
        """Get the log-posterior (up to an additive constant) for all chains.

        Each element of the returned array is the log-posterior for
        a chain. Optional parameter `inc_warmup` indicates whether to
        include the warmup period.
        """
        self._verify_has_samples()
        llp =  [ch['chains']['lp__'] for ch in self.sim['samples']]
        return llp if inc_warmup else [x[warmup2:] for x, warmup2 in zip(llp, self.sim['warmup2'])]

    def get_sampler_params(self, inc_warmup=True):
        """Obtain the parameters used for the sampler such as `stepsize` and
        `treedepth`. The results are returned as a list, each element of which
        is an OrderedDict a chain. The dictionary has number of elements
        corresponding to the number of parameters used in the sampler. Optional
        parameter `inc_warmup` indicates whether to include the warmup period.
        """
        self._verify_has_samples()
        ldf = [OrderedDict(zip(ch['sampler_param_names'], np.array(ch['sampler_params']))) for ch in self.sim['samples']]
        if inc_warmup:
            return ldf
        else:
            for d, warmup2 in zip(ldf, self.sim['warmup2']):
                for key in d:
                    d[key] = d[key][warmup2:]
            return ldf

    def get_posterior_mean(self):
        """Get the posterior mean for all parameters

        Returns
        -------
        means : array of shape (num_parameters, num_chains)
            Order of parameters is given by self.model_pars or self.flatnames
            if parameters of interest include non-scalar parameters. An additional
            column for mean lp__ is also included.
        """
        self._verify_has_samples()
        fnames = self.flatnames
        mean_pars = np.array([ch['mean_pars'] for ch in self.sim['samples']])
        mean_lp__ = np.array([ch['mean_lp__'] for ch in self.sim['samples']])
        mean_pars = np.column_stack(mean_pars)
        assert len(fnames) == len(mean_pars)
        m = np.row_stack([mean_pars, mean_lp__])
        return m

    def constrain_pars(self, np.ndarray[double, ndim=1, mode="c"] upar not None):
        """Transform parameters from unconstrained space to defined support"""
        cdef vector[double] constrained
        constrained = self.thisptr.constrain_pars(upar)
        return np.asarray(constrained)

    def unconstrain_pars(self, par):
        """Transform parameters from defined support to unconstrained space"""
        cdef vector[double] unconstrained
        data_r, data_i = pystan.misc._split_data(par)
        cdef vars_r_t vars_r = _dict_to_vars_r(data_r)
        cdef vars_i_t vars_i = _dict_to_vars_i(data_i)
        unconstrained = self.thisptr.unconstrain_pars(vars_r, vars_i)
        return np.asarray(unconstrained)

    def get_seed(self):
        return self.stan_args[0]['seed']

    def get_inits(self):
        return self.inits

    def get_stancode(self):
        return self.stanmodel.model_code

    def get_stanmodel(self):
        return self.stanmodel

    def get_stepsize(self):
        """Parse stepsize from fit object

        Parameters
        ----------
        fit : StanFit4Model

        Returns
        -------
        list
            Returns an empty list if step sizes
            are not found in ``fit.get_adaptation_info``.
        """
        return pystan.misc.get_stepsize(fit=self)

    def get_inv_metric(self, as_dict=False):
        """Parse inverse metrics from the fit object

        Parameters
        ----------
        as_dict : bool, optional


        Returns
        -------
        list
            Returns an empty list if inverse metric
            is not found in ``fit.get_adaptation_info()``.
        """
        return pystan.misc.get_inv_metric(fit=self, as_dict=as_dict)

    def get_last_position(self, warmup=False):
        """Parse last position from fit object

        Parameters
        ----------
        warmup : bool
            If True, returns the last warmup position, when warmup has been done.
            Otherwise function returns the first sample position.

        Returns
        -------
        list
            list contains a dictionary of last draw from each chain.
        """
        return pystan.misc.get_last_position(fit=self, warmup=warmup)

    def to_dataframe(self, pars=None, permuted=False, dtypes=None, inc_warmup=False, diagnostics=True, header=True):
        """Extract samples as a pandas dataframe for different parameters.

        Parameters
        ----------
        pars : {str, sequence of str}
            parameter (or quantile) name(s).
        permuted : bool, default False
            If True, returned samples are permuted.
            If inc_warmup is True, warmup samples have negative order.
        dtypes : dict
            datatype of parameter(s).
            If nothing is passed, float will be used for all parameters.
        inc_warmup : bool
            If True, warmup samples are kept; otherwise they are
            discarded.
        diagnostics : bool
            If True, include hmc diagnostics in dataframe.
        header : bool
            If True, include header columns.


        Returns
        -------
        df : pandas dataframe
            Returned dataframe contains: [header_df]|[draws_df]|[diagnostics_df],
            where all groups are optional.
            To exclude draws_df use `pars=[]`.


        Note
        ----
        Unlike default in extract (`permuted=True`)
        `.to_dataframe` method returns non-permuted samples (`permuted=False`) with diagnostics params included.
        """
        return pystan.misc.to_dataframe(fit=self, pars=pars, permuted=permuted, dtypes=dtypes, inc_warmup=inc_warmup, diagnostics=diagnostics, header=header)


    # FIXME: when this is a normal Python class one can use @property instead
    # of this special Cython syntax.
    property flatnames:

        def __get__(self):
            # NOTE: RStan rewrites the C++ function get_all_flatnames in R (in misc.R).
            # PyStan exposes and calls the C++ function directly.
            cdef vector[string] fnames
            names = [n.encode('ascii') for n in self.model_pars]
            get_all_flatnames(names, self.par_dims, fnames, col_major=True)
            return [n.decode('ascii') for n in fnames]


    # "private" Python methods

    def _verify_has_samples(self):
        if self.mode == 1:
            msg = "Stan model {} is of mode 'test_grad';\n" \
                "sampling is not conducted."
            raise AttributeError(msg.format(self.model_name))
        elif self.mode == 2 or self.sim.get('samples') is None:
            msg = "Stan model {} does not contain samples."
            raise AttributeError(msg.format(self.model_name))

    def _update_param_oi(self, pars):
        pars_bytes = [n.encode('ascii') for n in pars]
        cdef vector[string] pars_ = pars_bytes
        cdef int ret = self.thisptr.update_param_oi(pars_)
        return ret

    def _get_param_names(self):
        cdef vector[string] param_names_bytes = self.thisptr.param_names()
        param_names = [n.decode('utf-8') for n in param_names_bytes]
        return param_names

    def _get_param_fnames_oi(self):
        cdef vector[string] param_fnames_bytes = self.thisptr.param_fnames_oi()
        param_fnames = [n.decode('utf-8') for n in param_fnames_bytes]
        return param_fnames

    def _get_param_names_oi(self):
        cdef vector[string] param_names_bytes = self.thisptr.param_names_oi()
        param_names = [n.decode('utf-8') for n in param_names_bytes]
        return param_names

    def _get_param_dims(self):
        cdef vector[vector[uint]] dims = self.thisptr.param_dims()
        dims_ = dims
        return dims_

    def _get_param_dims_oi(self):
        cdef vector[vector[uint]] dims = self.thisptr.param_dims_oi()
        dims_ = dims
        return dims_

    def constrained_param_names(self):
        cdef vector[string] param_names_bytes = self.thisptr.constrained_param_names(False, False)
        param_names = [n.decode('utf-8') for n in param_names_bytes]
        return param_names

    def unconstrained_param_names(self):
        cdef vector[string] param_names_bytes = self.thisptr.unconstrained_param_names(False, False)
        param_names = [n.decode('utf-8') for n in param_names_bytes]
        return param_names

    def _call_sampler(self, dict args, pars_oi=None):
        return _call_sampler(self.data, args, pars_oi=pars_oi)



import unittest
from pystan.tests.helper import get_model


import pystan


class Test8Schools(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        schools_code = """\
            data {
                int<lower=0> J; // number of schools
                real y[J]; // estimated treatment effects
                real<lower=0> sigma[J]; // s.e. of effect estimates
            }
            parameters {
                real mu;
                real<lower=0> tau;
                real eta[J];
            }
            transformed parameters {
                real theta[J];
                for (j in 1:J)
                    theta[j] = mu + tau * eta[j];
            }
            model {
                eta ~ normal(0, 1);
                y ~ normal(theta, sigma);
            }"""

        cls.schools_dat = {'J': 8,
                           'y': [28,  8, -3,  7, -1,  1, 18, 12],
                           'sigma': [15, 10, 16, 11,  9, 11, 10, 18]}

        cls.model = get_model("schools_model", schools_code)
        #cls.model pystan.StanModel(model_code=schools_code)

    def test_8schools_pars(self):
        model = self.model
        data = self.schools_dat
        pars = ['mu']  # list of parameters
        fit = model.sampling(data=data, pars=pars, iter=100)
        self.assertEqual(len(fit.sim['pars_oi']), len(fit.sim['dims_oi']))
        self.assertEqual(len(fit.extract()), 2)
        self.assertIn('mu', fit.extract())
        self.assertRaises(ValueError, fit.extract, 'theta')
        self.assertIsNotNone(fit.extract(permuted=False))

        pars = ['eta']
        fit = model.sampling(data=data, pars=pars, iter=100)
        self.assertEqual(len(fit.extract()), 2)
        self.assertIn('eta', fit.extract())
        self.assertRaises(ValueError, fit.extract, 'theta')
        self.assertIsNotNone(fit.extract(permuted=False))

        pars = ['mu', 'eta']
        fit = model.sampling(data=data, pars=pars, iter=100)
        self.assertEqual(len(fit.extract()), 3)
        self.assertIn('mu', fit.extract())
        self.assertIn('eta', fit.extract())
        self.assertRaises(ValueError, fit.extract, 'theta')
        self.assertIsNotNone(fit.extract(permuted=False))

    def test_8schools_pars_bare(self):
        model = self.model
        data = self.schools_dat
        pars = 'mu'  # bare string
        fit = model.sampling(data=data, pars=pars, iter=100)
        self.assertEqual(len(fit.extract()), 2)
        self.assertIn('mu', fit.extract())
        self.assertRaises(ValueError, fit.extract, 'theta')
        self.assertIsNotNone(fit.extract(permuted=False))

        pars = 'eta'  # bare string
        fit = model.sampling(data=data, pars=pars)
        self.assertEqual(len(fit.extract()), 2)
        self.assertIn('eta', fit.extract())
        self.assertRaises(ValueError, fit.extract, 'theta')
        self.assertIsNotNone(fit.extract(permuted=False))

    def test_8schools_bad_pars(self):
        model = self.model
        data = self.schools_dat
        pars = ['mu', 'missing']  # 'missing' is not in the model
        self.assertRaises(ValueError, model.sampling,
                          data=data, pars=pars, iter=100)


class TestParsLabels(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        model_code = '''\
            parameters {
              real a;
              real b;
              real c;
              real d;
            }
            model {
              a ~ normal(0, 1);
              b ~ normal(-19, 1);
              c ~ normal(5, 1);
              d ~ normal(11, 1);
            }
        '''
        cls.model = pystan.StanModel(model_code=model_code)

    def test_pars_single(self):
        '''Make sure labels are not getting switched'''
        fit = self.model.sampling(iter=100, chains=1, n_jobs=1, pars=['d'])
        self.assertGreater(fit.extract()['d'].mean(), 7)

    def test_pars_single_chains(self):
        '''Make sure labels are not getting switched'''
        fit = self.model.sampling(iter=100, chains=2, n_jobs=2, pars=['d'])
        self.assertGreater(fit.extract()['d'].mean(), 7)

    def test_pars_labels(self):
        '''Make sure labels are not getting switched'''
        fit = self.model.sampling(iter=100, chains=2, n_jobs=2, pars=['a', 'd'])
        extr = fit.extract()
        a = extr['a'].mean()
        d = extr['d'].mean()
        self.assertTrue('b' not in extr)
        self.assertTrue('c' not in extr)
        self.assertGreater(d, a)



import unittest

import numpy as np

import pystan


class TestOptimizingExample(unittest.TestCase):
    """Test optimizing example from documentation"""

    @classmethod
    def setUpClass(cls):
        ocode = """
        data {
            int<lower=1> N;
            real y[N];
        }
        parameters {
            real mu;
        }
        model {
            y ~ normal(mu, 1);
        }
        """

        cls.sm = pystan.StanModel(model_code=ocode)

    def test_optimizing(self):
        sm = self.sm
        np.random.seed(3)
        y2 = np.random.normal(size=20)
        op = sm.optimizing(data=dict(y=y2, N=len(y2)))
        self.assertAlmostEqual(op['mu'], np.mean(y2))



from collections import OrderedDict
import os
import tempfile
import unittest

import numpy as np

from pystan import misc
from pystan._compat import PY2
from pystan.constants import MAX_UINT


def is_valid_seed(seed):
    return isinstance(seed, int) and seed >= 0 and seed <= MAX_UINT


class TestMisc(unittest.TestCase):

    def test_pars_total_indexes(self):
        pars_oi = ['mu', 'tau', 'eta', 'theta', 'lp__']
        dims_oi = [[], [], [8], [8], []]
        fnames_oi = ['mu', 'tau', 'eta[1]', 'eta[2]', 'eta[3]', 'eta[4]', 'eta[5]',
                     'eta[6]', 'eta[7]', 'eta[8]', 'theta[1]', 'theta[2]',
                     'theta[3]', 'theta[4]', 'theta[5]', 'theta[6]', 'theta[7]',
                     'theta[8]', 'lp__']
        pars = ['mu', 'tau', 'eta', 'theta', 'lp__']
        rslt = misc._pars_total_indexes(pars_oi, dims_oi, fnames_oi, pars)
        self.assertEqual(rslt['mu'], (0,))
        self.assertEqual(rslt['tau'], (1,))
        self.assertEqual(rslt['eta'], (2, 3, 4, 5, 6, 7, 8, 9))
        self.assertEqual(rslt['theta'], (10, 11, 12, 13, 14, 15, 16, 17))
        self.assertEqual(rslt['lp__'], (18,))

    def test_par_vector2dict(self):
        v = [0, 1, -1, -1, 0, 1, -1, -2]
        pars = ['alpha', 'beta']
        dims = [[2, 3], [2]]
        rslt = misc._par_vector2dict(v, pars, dims)
        np.testing.assert_equal(rslt['alpha'], np.array([0, 1, -1, -1, 0, 1]).reshape(dims[0], order='F'))
        np.testing.assert_equal(rslt['beta'], np.array([-1, -2]))

    def test_check_seed(self):
        self.assertEqual(misc._check_seed('10'), 10)
        self.assertEqual(misc._check_seed(10), 10)
        self.assertEqual(misc._check_seed(10.5), 10)
        self.assertTrue(is_valid_seed(misc._check_seed(np.random.RandomState())))
        self.assertTrue(is_valid_seed(misc._check_seed(-1)))

    def test_check_seed_gt_max_uint(self):
        assertRaisesRegex = self.assertRaisesRegexp if PY2 else self.assertRaisesRegex
        with assertRaisesRegex(ValueError, '`seed` is too large'):
            misc._check_seed(MAX_UINT + 1)

    def test_stan_rdump(self):
        data = OrderedDict(x=1, y=0, z=[1, 2, 3], Phi=np.array([[1, 2], [3, 4]]))
        tempdir = tempfile.mkdtemp()
        filename = os.path.join(tempdir, 'test_rdump.rdump')
        misc.stan_rdump(data, filename)
        data_recovered = misc.read_rdump(filename)
        np.testing.assert_equal(data, data_recovered)

        data = OrderedDict(x=True, y=np.array([True, False]))
        data_expected = OrderedDict(x=1, y=np.array([1, 0]))
        misc.stan_rdump(data, filename)
        data_recovered = misc.read_rdump(filename)
        np.testing.assert_equal(data_recovered, data_expected)

        data = OrderedDict(x='foo')
        np.testing.assert_raises(ValueError, misc.stan_rdump, data, filename)

        data = OrderedDict(new=3)
        np.testing.assert_raises(ValueError, misc.stan_rdump, data, filename)

    def test_stan_rdump_array(self):
        y = np.asarray([[1, 3, 5], [2, 4, 6]])
        data = {'y': y}
        expected = 'y <-\nstructure(c(1, 2, 3, 4, 5, 6), .Dim = c(2, 3))\n'
        np.testing.assert_equal(misc._dict_to_rdump(data), expected)

    def test_stan_read_rdump_array(self):
        """
        For reference:
        > structure(c(1, 2, 3, 4, 5, 6), .Dim = c(2, 3))
            [,1] [,2] [,3]
        [1,]    1    3    5
        [2,]    2    4    6
        """
        # make this into some sort of stream
        rdump = 'y <-\nstructure(c(1, 2, 3, 4, 5, 6), .Dim = c(2, 3))\n'
        tempdir = tempfile.mkdtemp()
        filename = os.path.join(tempdir, 'test_rdump.rdump')
        with open(filename, 'w') as f:
            f.write(rdump)
        d = misc.read_rdump(filename)
        y = np.asarray([[1, 3, 5], [2, 4, 6]])
        np.testing.assert_equal(d['y'], y)

    def test_rstan_read_rdump(self):
        a = np.array([1, 3, 5])
        b = np.arange(10).reshape((-1, 2))
        c = np.arange(18).reshape([2, 3, 3])
        data = dict(a=a, b=b, c=c)
        tempdir = tempfile.mkdtemp()
        filename = os.path.join(tempdir, 'test_rdump.rdump')
        misc.stan_rdump(data, filename)
        d = misc.read_rdump(filename)
        np.testing.assert_equal(d['a'], a)
        np.testing.assert_equal(d['b'], b)
        np.testing.assert_equal(d['c'], c)

    def test_remove_empty_pars(self):
        pars = ('alpha', 'beta', 'gamma', 'eta', 'xi')
        dims = [[], (2,), (2, 4), 0, (2, 0)]
        # NOTE: np.prod([]) -> 1
        np.testing.assert_equal(misc._remove_empty_pars(pars[0:1], pars, dims), pars[0:1])
        np.testing.assert_equal(misc._remove_empty_pars(pars[0:3], pars, dims), pars[0:3])
        np.testing.assert_equal(misc._remove_empty_pars(pars[0:4], pars, dims), pars[0:3])
        np.testing.assert_equal(misc._remove_empty_pars(['beta[1]'], pars, dims), ['beta[1]'])
        np.testing.assert_equal(misc._remove_empty_pars(['eta'], pars, dims), [])

    def test_format_number_si(self):
        np.testing.assert_equal(misc._format_number_si(-12345, 2), '-1.2e4')
        np.testing.assert_equal(misc._format_number_si(123456, 2), '1.2e5')
        np.testing.assert_equal(misc._format_number_si(-123456, 2), '-1.2e5')
        np.testing.assert_equal(misc._format_number_si(-123456.393, 2), '-1.2e5')
        np.testing.assert_equal(misc._format_number_si(1234567, 2), '1.2e6')
        np.testing.assert_equal(misc._format_number_si(0.12, 2), '1.2e-1')
        np.testing.assert_equal(misc._format_number_si(3.32, 2), '3.3e0')

    def test_format_number(self):
        np.testing.assert_equal(misc._format_number(-1234, 2, 6), '-1234')
        np.testing.assert_equal(misc._format_number(-12345, 2, 6), '-12345')
        np.testing.assert_equal(misc._format_number(123456, 2, 6), '123456')
        np.testing.assert_equal(misc._format_number(-123456, 2, 6), '-1.2e5')
        np.testing.assert_equal(misc._format_number(1234567, 2, 6), '1.2e6')
        np.testing.assert_equal(misc._format_number(0.12, 2, 6), '0.12')
        np.testing.assert_equal(misc._format_number(3.32, 2, 6), '3.32')
        np.testing.assert_equal(misc._format_number(3.3, 2, 6), '3.3')
        np.testing.assert_equal(misc._format_number(3.32, 2, 6), '3.32')
        np.testing.assert_equal(misc._format_number(3.323945, 2, 6), '3.32')
        np.testing.assert_equal(misc._format_number(-0.0003434, 2, 6), '-3.4e-4')
        np.testing.assert_equal(misc._format_number(1654.25, 2, 6), '1654.2')
        np.testing.assert_equal(misc._format_number(-1654.25, 2, 6), '-1654')
        np.testing.assert_equal(misc._format_number(-165.25, 2, 6), '-165.2')
        np.testing.assert_equal(misc._format_number(9.94598693e-01, 2, 6), '0.99')
        np.testing.assert_equal(misc._format_number(-9.94598693e-01, 2, 6), '-0.99')

    def test_format_number_inf(self):
        np.testing.assert_equal(misc._format_number_si(float('inf'), 2), 'inf')
        np.testing.assert_equal(misc._format_number_si(float('-inf'), 2), '-inf')

    def test_array_to_table(self):
        rownames = np.array(['alpha', 'beta[0]', 'beta[1]', 'beta[2]', 'beta[3]', 'sigma', 'lp__'])
        colnames = ('mean', 'se_mean', 'sd', '2.5%', '25%', '50%', '75%', '97.5%', 'n_eff', 'Rhat')
        arr = np.array([[-1655,   625,  1654, -4378, -3485,  -927,   -27,    -9,     7, 2],
                        [  -47,    59,   330,  -770,   -69,    -2,     2,  1060,    31, 1],
                        [  -18,    70,   273,  -776,   -30,     0,    35,   608,    15, 1],
                        [    8,    38,   230,  -549,   -26,     0,    17,   604,    36, 1],
                        [   23,    32,   303,  -748,   -34,     0,    49,   751,    92, 1],
                        [  405,   188,   974,     0,     4,    73,   458,  2203,    27, 1],
                        [  -13,     3,     8,   -23,   -20,   -15,    -5,     1,     8, 2]])
        n_digits = 2
        result = misc._array_to_table(arr, rownames, colnames, n_digits)
        desired = "alpha    -1655     625   1654  -4378  -3485   -927    -27     -9      7      2"
        desired_py2 = "alpha    -1655   625.0 1654.0  -4378  -3485 -927.0  -27.0   -9.0    7.0    2.0"
        # round() behaves differently in Python 3
        if PY2:
            np.testing.assert_equal(result.split('\n')[1], desired_py2)
        else:
            np.testing.assert_equal(result.split('\n')[1], desired)
        arr = np.array([[-1655325, 625.25,  1654.25, -4378.25, -3485.25,  -927.25, -27.25,    -9.25,   7.25, 2],
                        [  -47.25,  59.25,   330.25,  -770.25,   -69.25,    -2.25,   2.25,  1060.25,  31.25, 1],
                        [  -18.25,  70.25,   273.25,  -776.25,   -30.25,     0.25,  35.25,   608.25,  15.25, 1],
                        [    8.25,  38.25,   230.25,  -549.25,   -26.25,     0.25,  17.25,   604.25,  36.25, 1],
                        [   23.25,  32.25,   303.25,  -748.25,   -34.25,     0.25,  49.25,   751.25,  92.25, 1],
                        [  405.25, 188.25,   974.25,     0.25,     4.25,    73.25, 458.25,  2203.25,  27.25, 1],
                        [  -13.25,   3.25,     8.25,   -23.25,   -20.25,   -15.25,  -5.25,     1.25,   8.25, 2]])
        result = misc._array_to_table(arr, rownames, colnames, n_digits)
        desired = "alpha   -1.7e6  625.25 1654.2  -4378  -3485 -927.2 -27.25  -9.25      7    2.0"
        desired_py2 = "alpha   -1.7e6  625.25 1654.2  -4378  -3485 -927.2 -27.25  -9.25    7.0    2.0"
        # round() behaves differently in Python 3
        if PY2:
            np.testing.assert_equal(result.split('\n')[1], desired_py2)
        else:
            np.testing.assert_equal(result.split('\n')[1], desired)



"""
mquantiles function from scipy
scipy commit: 195569e91b0708aae26ea74581d03d3ca279117e

:author: Pierre GF Gerard-Marchant
:contact: pierregm_at_uga_edu

scipy is distributed under a BSD license

"""

import numpy as np
import numpy.ma as ma
from numpy.ma import masked, nomask

def mquantiles(a, prob=list([.25,.5,.75]), alphap=.4, betap=.4, axis=None,
               limit=()):
    """
    Computes empirical quantiles for a data array.

    Samples quantile are defined by ``Q(p) = (1-gamma)*x[j] + gamma*x[j+1]``,
    where ``x[j]`` is the j-th order statistic, and gamma is a function of
    ``j = floor(n*p + m)``, ``m = alphap + p*(1 - alphap - betap)`` and
    ``g = n*p + m - j``.

    Reinterpreting the above equations to compare to **R** lead to the
    equation: ``p(k) = (k - alphap)/(n + 1 - alphap - betap)``

    Typical values of (alphap,betap) are:
        - (0,1)    : ``p(k) = k/n`` : linear interpolation of cdf
          (**R** type 4)
        - (.5,.5)  : ``p(k) = (k - 1/2.)/n`` : piecewise linear function
          (**R** type 5)
        - (0,0)    : ``p(k) = k/(n+1)`` :
          (**R** type 6)
        - (1,1)    : ``p(k) = (k-1)/(n-1)``: p(k) = mode[F(x[k])].
          (**R** type 7, **R** default)
        - (1/3,1/3): ``p(k) = (k-1/3)/(n+1/3)``: Then p(k) ~ median[F(x[k])].
          The resulting quantile estimates are approximately median-unbiased
          regardless of the distribution of x.
          (**R** type 8)
        - (3/8,3/8): ``p(k) = (k-3/8)/(n+1/4)``: Blom.
          The resulting quantile estimates are approximately unbiased
          if x is normally distributed
          (**R** type 9)
        - (.4,.4)  : approximately quantile unbiased (Cunnane)
        - (.35,.35): APL, used with PWM

    Parameters
    ----------
    a : array_like
        Input data, as a sequence or array of dimension at most 2.
    prob : array_like, optional
        List of quantiles to compute.
    alphap : float, optional
        Plotting positions parameter, default is 0.4.
    betap : float, optional
        Plotting positions parameter, default is 0.4.
    axis : int, optional
        Axis along which to perform the trimming.
        If None (default), the input array is first flattened.
    limit : tuple
        Tuple of (lower, upper) values.
        Values of `a` outside this open interval are ignored.

    Returns
    -------
    mquantiles : MaskedArray
        An array containing the calculated quantiles.

    Notes
    -----
    This formulation is very similar to **R** except the calculation of
    ``m`` from ``alphap`` and ``betap``, where in **R** ``m`` is defined
    with each type.

    References
    ----------
    .. [1] *R* statistical software at http://www.r-project.org/

    Examples
    --------
    >>> from scipy.stats.mstats import mquantiles
    >>> a = np.array([6., 47., 49., 15., 42., 41., 7., 39., 43., 40., 36.])
    >>> mquantiles(a)
    array([ 19.2,  40. ,  42.8])

    Using a 2D array, specifying axis and limit.

    >>> data = np.array([[   6.,    7.,    1.],
                         [  47.,   15.,    2.],
                         [  49.,   36.,    3.],
                         [  15.,   39.,    4.],
                         [  42.,   40., -999.],
                         [  41.,   41., -999.],
                         [   7., -999., -999.],
                         [  39., -999., -999.],
                         [  43., -999., -999.],
                         [  40., -999., -999.],
                         [  36., -999., -999.]])
    >>> mquantiles(data, axis=0, limit=(0, 50))
    array([[ 19.2 ,  14.6 ,   1.45],
           [ 40.  ,  37.5 ,   2.5 ],
           [ 42.8 ,  40.05,   3.55]])

    >>> data[:, 2] = -999.
    >>> mquantiles(data, axis=0, limit=(0, 50))
    masked_array(data =
     [[19.2 14.6 --]
     [40.0 37.5 --]
     [42.8 40.05 --]],
                 mask =
     [[False False  True]
      [False False  True]
      [False False  True]],
           fill_value = 1e+20)

    """
    def _quantiles1D(data,m,p):
        x = np.sort(data.compressed())
        n = len(x)
        if n == 0:
            return ma.array(np.empty(len(p), dtype=float), mask=True)
        elif n == 1:
            return ma.array(np.resize(x, p.shape), mask=nomask)
        aleph = (n*p + m)
        k = np.floor(aleph.clip(1, n-1)).astype(int)
        gamma = (aleph-k).clip(0,1)
        return (1.-gamma)*x[(k-1).tolist()] + gamma*x[k.tolist()]

    # Initialization & checks ---------
    data = ma.array(a, copy=False)
    if data.ndim > 2:
        raise TypeError("Array should be 2D at most !")
    #
    if limit:
        condition = (limit[0] < data) & (data < limit[1])
        data[~condition.filled(True)] = masked
    #
    p = np.array(prob, copy=False, ndmin=1)
    m = alphap + p*(1.-alphap-betap)
    # Computes quantiles along axis (or globally)
    if (axis is None):
        return _quantiles1D(data, m, p)
    return ma.apply_along_axis(_quantiles1D, axis, data, m, p)



"""PyStan utility functions

These functions validate and organize data passed to and from the
classes and functions defined in the file `stan_fit.hpp` and wrapped
by the Cython file `stan_fit.pxd`.

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2013-2015, PyStan developers
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------

# REF: rstan/rstan/R/misc.R

from __future__ import unicode_literals, division
from pystan._compat import PY2, string_types

from collections import OrderedDict
if PY2:
    from collections import Callable, Iterable, Sequence
else:
    from collections.abc import Callable, Iterable, Sequence
import inspect
import io
import itertools
import logging
import math
from numbers import Number
import os
import random
import re
import sys
import shutil
import tempfile
import time

import numpy as np
try:
    from scipy.stats.mstats import mquantiles
except ImportError:
    from pystan.external.scipy.mstats import mquantiles

import pystan.chains
import pystan._misc
from pystan.constants import (MAX_UINT, sampling_algo_t, optim_algo_t,
                              variational_algo_t, sampling_metric_t, stan_args_method_t)

logger = logging.getLogger('pystan')


def stansummary(fit, pars=None, probs=(0.025, 0.25, 0.5, 0.75, 0.975), digits_summary=2):
    """
    Summary statistic table.

    Parameters
    ----------
    fit : StanFit4Model object
    pars : str or sequence of str, optional
        Parameter names. By default use all parameters
    probs : sequence of float, optional
        Quantiles. By default, (0.025, 0.25, 0.5, 0.75, 0.975)
    digits_summary : int, optional
        Number of significant digits. By default, 2

    Returns
    -------
    summary : string
        Table includes mean, se_mean, sd, probs_0, ..., probs_n, n_eff and Rhat.

    Examples
    --------
    >>> model_code = 'parameters {real y;} model {y ~ normal(0,1);}'
    >>> m = StanModel(model_code=model_code, model_name="example_model")
    >>> fit = m.sampling()
    >>> print(stansummary(fit))
    Inference for Stan model: example_model.
    4 chains, each with iter=2000; warmup=1000; thin=1;
    post-warmup draws per chain=1000, total post-warmup draws=4000.

           mean se_mean     sd   2.5%    25%    50%    75%  97.5%  n_eff   Rhat
    y      0.01    0.03    1.0  -2.01  -0.68   0.02   0.72   1.97   1330    1.0
    lp__   -0.5    0.02   0.68  -2.44  -0.66  -0.24  -0.05-5.5e-4   1555    1.0

    Samples were drawn using NUTS at Thu Aug 17 00:52:25 2017.
    For each parameter, n_eff is a crude measure of effective sample size,
    and Rhat is the potential scale reduction factor on split chains (at
    convergence, Rhat=1).
    """
    if fit.mode == 1:
        return "Stan model '{}' is of mode 'test_grad';\n"\
               "sampling is not conducted.".format(fit.model_name)
    elif fit.mode == 2:
        return "Stan model '{}' does not contain samples.".format(fit.model_name)

    n_kept = [s - w for s, w in zip(fit.sim['n_save'], fit.sim['warmup2'])]
    header = "Inference for Stan model: {}.\n".format(fit.model_name)
    header += "{} chains, each with iter={}; warmup={}; thin={}; \n"
    header = header.format(fit.sim['chains'], fit.sim['iter'], fit.sim['warmup'],
                           fit.sim['thin'], sum(n_kept))
    header += "post-warmup draws per chain={}, total post-warmup draws={}.\n\n"
    header = header.format(n_kept[0], sum(n_kept))
    footer = "\n\nSamples were drawn using {} at {}.\n"\
        "For each parameter, n_eff is a crude measure of effective sample size,\n"\
        "and Rhat is the potential scale reduction factor on split chains (at \n"\
        "convergence, Rhat=1)."
    sampler = fit.sim['samples'][0]['args']['sampler_t']
    date = fit.date.strftime('%c')  # %c is locale's representation
    footer = footer.format(sampler, date)
    s = _summary(fit, pars, probs)
    body = _array_to_table(s['summary'], s['summary_rownames'],
                           s['summary_colnames'], digits_summary)
    return header + body + footer

def _print_stanfit(fit, pars=None, probs=(0.025, 0.25, 0.5, 0.75, 0.975), digits_summary=2):
    # warning added in PyStan 2.17.0
    logger.warning('Function `_print_stanfit` is deprecated and will be removed in a future version. '\
                  'Use `stansummary` instead.', DeprecationWarning)
    return stansummary(fit, pars=pars, probs=probs, digits_summary=digits_summary)

def _array_to_table(arr, rownames, colnames, n_digits):
    """Print an array with row and column names

    Example:
                  mean se_mean  sd 2.5%  25%  50%  75% 97.5% n_eff Rhat
        beta[1,1]  0.0     0.0 1.0 -2.0 -0.7  0.0  0.7   2.0  4000    1
        beta[1,2]  0.0     0.0 1.0 -2.1 -0.7  0.0  0.7   2.0  4000    1
        beta[2,1]  0.0     0.0 1.0 -2.0 -0.7  0.0  0.7   2.0  4000    1
        beta[2,2]  0.0     0.0 1.0 -1.9 -0.6  0.0  0.7   2.0  4000    1
        lp__      -4.2     0.1 2.1 -9.4 -5.4 -3.8 -2.7  -1.2   317    1
    """
    assert arr.shape == (len(rownames), len(colnames))
    rownames_maxwidth = max(len(n) for n in rownames)
    max_col_width = 7
    min_col_width = 5
    max_col_header_num_width = [max(max_col_width, max(len(n) + 1, min_col_width)) for n in colnames]
    rows = []
    for row in arr:
        row_nums = []
        for j, (num, width) in enumerate(zip(row, max_col_header_num_width)):
            if colnames[j] == "n_eff":
                num = int(round(num, 0)) if not np.isnan(num) else num
            num = _format_number(num, n_digits, max_col_width - 1)
            row_nums.append(num)
            if len(num) + 1 > max_col_header_num_width[j]:
                max_col_header_num_width[j] = len(num) + 1
        rows.append(row_nums)
    widths = [rownames_maxwidth] + max_col_header_num_width
    header = '{:>{width}}'.format('', width=widths[0])
    for name, width in zip(colnames, widths[1:]):
        header += '{name:>{width}}'.format(name=name, width=width)
    lines = [header]
    for rowname, row in zip(rownames, rows):
        line = '{name:{width}}'.format(name=rowname, width=widths[0])
        for j, (num, width) in enumerate(zip(row, widths[1:])):
            line += '{num:>{width}}'.format(num=num, width=width)
        lines.append(line)
    return '\n'.join(lines)


def _number_width(n):
    """Calculate the width in characters required to print a number

    For example, -1024 takes 5 characters. -0.034 takes 6 characters.
    """
    return len(str(n))


def _format_number_si(num, n_signif_figures):
    """Format a number using scientific notation to given significant figures"""
    if math.isnan(num) or math.isinf(num):
        return str(num)
    leading, exp = '{:E}'.format(num).split('E')
    leading = round(float(leading), n_signif_figures - 1)
    exp = exp[:1] + exp[2:] if exp[1] == '0' else exp
    formatted = '{}e{}'.format(leading, exp.lstrip('+'))
    return formatted


def _format_number(num, n_signif_figures, max_width):
    """Format a number as a string while obeying space constraints.

    `n_signif_figures` is the minimum number of significant figures expressed
    `max_width` is the maximum width in characters allowed
    """
    if max_width < 6:
        raise NotImplementedError("Guaranteed formatting in fewer than 6 characters not supported.")
    if math.isnan(num) or math.isinf(num):
        return str(num)
    # add 0.5 to prevent log(0) errors; only affects n_digits calculation for num > 0
    n_digits = lambda num: math.floor(math.log10(abs(num) + 0.5)) + 1
    if abs(num) > 10**-n_signif_figures and n_digits(num) <= max_width - n_signif_figures:
        return str(round(num, n_signif_figures))[:max_width].rstrip('.')
    elif _number_width(num) <= max_width:
        if n_digits(num) >= n_signif_figures:
            # the int() is necessary for consistency between Python 2 and 3
            return str(int(round(num)))
        else:
            return str(num)
    else:
        return _format_number_si(num, n_signif_figures)


def _summary(fit, pars=None, probs=None, **kwargs):
    """Summarize samples (compute mean, SD, quantiles) in all chains.

    REF: stanfit-class.R summary method

    Parameters
    ----------
    fit : StanFit4Model object
    pars : str or sequence of str, optional
        Parameter names. By default use all parameters
    probs : sequence of float, optional
        Quantiles. By default, (0.025, 0.25, 0.5, 0.75, 0.975)

    Returns
    -------
    summaries : OrderedDict of array
        Array indexed by 'summary' has dimensions (num_params, num_statistics).
        Parameters are unraveled in *row-major order*. Statistics include: mean,
        se_mean, sd, probs_0, ..., probs_n, n_eff, and Rhat. Array indexed by
        'c_summary' breaks down the statistics by chain and has dimensions
        (num_params, num_statistics_c_summary, num_chains). Statistics for
        `c_summary` are the same as for `summary` with the exception that
        se_mean, n_eff, and Rhat are absent. Row names and column names are
        also included in the OrderedDict.
    """
    if fit.mode == 1:
        msg = "Stan model {} is of mode 'test_grad'; sampling is not conducted."
        msg = msg.format(fit.model_name)
        raise ValueError(msg)
    elif fit.mode == 2:
        msg = "Stan model {} contains no samples.".format(fit.model_name)
        raise ValueError(msg)

    if fit.sim['n_save'] == fit.sim['warmup2']:
        msg = "Stan model {} contains no samples.".format(fit.model_name)
        raise ValueError(msg)

    # rstan checks for cached summaries here

    if pars is None:
        pars = fit.sim['pars_oi']
    elif isinstance(pars, string_types):
        pars = [pars]
    pars = _remove_empty_pars(pars, fit.sim['pars_oi'], fit.sim['dims_oi'])

    if probs is None:
        probs = (0.025, 0.25, 0.5, 0.75, 0.975)
    ss = _summary_sim(fit.sim, pars, probs)
    # TODO: include sem, ess and rhat: ss['ess'], ss['rhat']
    s1 = np.column_stack([ss['msd'][:, 0], ss['sem'], ss['msd'][:, 1], ss['quan'], ss['ess'], ss['rhat']])
    s1_rownames = ss['c_msd_names']['parameters']
    s1_colnames = ((ss['c_msd_names']['stats'][0],) + ('se_mean',) +
                   (ss['c_msd_names']['stats'][1],) + ss['c_quan_names']['stats'] +
                   ('n_eff', 'Rhat'))
    s2 = _combine_msd_quan(ss['c_msd'], ss['c_quan'])
    s2_rownames = ss['c_msd_names']['parameters']
    s2_colnames = ss['c_msd_names']['stats'] + ss['c_quan_names']['stats']
    return OrderedDict(summary=s1, c_summary=s2,
                       summary_rownames=s1_rownames,
                       summary_colnames=s1_colnames,
                       c_summary_rownames=s2_rownames,
                       c_summary_colnames=s2_colnames)


def _combine_msd_quan(msd, quan):
    """Combine msd and quantiles in chain summary

    Parameters
    ----------
    msd : array of shape (num_params, 2, num_chains)
       mean and sd for chains
    cquan : array of shape (num_params, num_quan, num_chains)
        quantiles for chains

    Returns
    -------
    msdquan : array of shape (num_params, 2 + num_quan, num_chains)
    """
    dim1 = msd.shape
    n_par, _, n_chains = dim1
    ll = []
    for i in range(n_chains):
        a1 = msd[:, :, i]
        a2 = quan[:, :, i]
        ll.append(np.column_stack([a1, a2]))
    msdquan = np.dstack(ll)
    return msdquan


def _summary_sim(sim, pars, probs):
    """Summarize chains together and separately

    REF: rstan/rstan/R/misc.R

    Parameters are unraveled in *column-major order*.

    Parameters
    ----------
    sim : dict
        dict from from a stanfit fit object, i.e., fit['sim']
    pars : Iterable of str
        parameter names
    probs : Iterable of probs
        desired quantiles

    Returns
    -------
    summaries : OrderedDict of array
        This dictionary contains the following arrays indexed by the keys
        given below:
        - 'msd' : array of shape (num_params, 2) with mean and sd
        - 'sem' : array of length num_params with standard error for the mean
        - 'c_msd' : array of shape (num_params, 2, num_chains)
        - 'quan' : array of shape (num_params, num_quan)
        - 'c_quan' : array of shape (num_params, num_quan, num_chains)
        - 'ess' : array of shape (num_params, 1)
        - 'rhat' : array of shape (num_params, 1)

    Note
    ----
    `_summary_sim` has the parameters in *column-major* order whereas `_summary`
    gives them in *row-major* order. (This follows RStan.)
    """
    # NOTE: this follows RStan rather closely. Some of the calculations here
    probs_len = len(probs)
    n_chains = len(sim['samples'])
    # tidx is a dict with keys that are parameters and values that are their
    # indices using column-major ordering
    tidx = _pars_total_indexes(sim['pars_oi'], sim['dims_oi'], sim['fnames_oi'], pars)
    tidx_colm = [tidx[par] for par in pars]
    tidx_colm = list(itertools.chain(*tidx_colm))  # like R's unlist()
    tidx_rowm = [tidx[par+'_rowmajor'] for par in pars]
    tidx_rowm = list(itertools.chain(*tidx_rowm))
    tidx_len = len(tidx_colm)
    lmsdq = [_get_par_summary(sim, i, probs) for i in tidx_colm]
    msd = np.row_stack([x['msd'] for x in lmsdq])
    quan = np.row_stack([x['quan'] for x in lmsdq])
    probs_str = tuple(["{:g}%".format(100*p) for p in probs])
    msd = msd.reshape(tidx_len, 2, order='F')
    quan = quan.reshape(tidx_len, probs_len, order='F')

    c_msd = np.row_stack([x['c_msd'] for x in lmsdq])
    c_quan = np.row_stack([x['c_quan'] for x in lmsdq])
    c_msd = c_msd.reshape(tidx_len, 2, n_chains, order='F')
    c_quan = c_quan.reshape(tidx_len, probs_len, n_chains, order='F')
    sim_attr_args = sim.get('args', None)
    if sim_attr_args is None:
        cids = list(range(n_chains))
    else:
        cids = [x['chain_id'] for x in sim_attr_args]

    c_msd_names = dict(parameters=np.asarray(sim['fnames_oi'])[tidx_colm],
                       stats=("mean", "sd"),
                       chains=tuple("chain:{}".format(cid) for cid in cids))
    c_quan_names = dict(parameters=np.asarray(sim['fnames_oi'])[tidx_colm],
                        stats=probs_str,
                        chains=tuple("chain:{}".format(cid) for cid in cids))
    ess_and_rhat = np.array([pystan.chains.ess_and_splitrhat(sim, n) for n in tidx_colm])
    ess, rhat = [arr.ravel() for arr in np.hsplit(ess_and_rhat, 2)]
    return dict(msd=msd, c_msd=c_msd, c_msd_names=c_msd_names, quan=quan,
                c_quan=c_quan, c_quan_names=c_quan_names,
                sem=msd[:, 1] / np.sqrt(ess), ess=ess, rhat=rhat,
                row_major_idx=tidx_rowm, col_major_idx=tidx_colm)


def _get_par_summary(sim, n, probs):
    """Summarize chains merged and individually

    Parameters
    ----------
    sim : dict from stanfit object
    n : int
        parameter index
    probs : iterable of int
        quantiles

    Returns
    -------
    summary : dict
       Dictionary containing summaries
    """
    # _get_samples gets chains for nth parameter
    ss = _get_samples(n, sim, inc_warmup=False)
    msdfun = lambda chain: (np.mean(chain), np.std(chain, ddof=1))
    qfun = lambda chain: mquantiles(chain, probs)
    c_msd = np.array([msdfun(s) for s in ss]).flatten()
    c_quan = np.array([qfun(s) for s in ss]).flatten()
    ass = np.asarray(ss).flatten()
    msd = np.asarray(msdfun(ass))
    quan = qfun(np.asarray(ass))
    return dict(msd=msd, quan=quan, c_msd=c_msd, c_quan=c_quan)


def _split_data(data):
    data_r = {}
    data_i = {}
    # data_r and data_i are going to be converted into C++ objects of
    # type: map<string, pair<vector<double>, vector<size_t>>> and
    # map<string, pair<vector<int>, vector<size_t>>> so prepare
    # them accordingly.
    for k, v in data.items():
        if np.issubdtype(np.asarray(v).dtype, np.integer):
            data_i.update({k.encode('utf-8'): np.asarray(v, dtype=int)})
        elif np.issubdtype(np.asarray(v).dtype, np.floating):
            data_r.update({k.encode('utf-8'): np.asarray(v, dtype=float)})
        else:
            msg = "Variable {} is neither int nor float nor list/array thereof"
            raise ValueError(msg.format(k))
    return data_r, data_i


def _config_argss(chains, iter, warmup, thin,
                  init, seed, sample_file, diagnostic_file, algorithm,
                  control, **kwargs):
    # After rstan/rstan/R/misc.R (config_argss)
    iter = int(iter)
    if iter < 1:
        raise ValueError("`iter` should be a positive integer.")
    thin = int(thin)
    if thin < 1 or thin > iter:
        raise ValueError("`thin should be a positive integer "
                         "less than `iter`.")
    warmup = max(0, int(warmup))
    if warmup > iter:
        raise ValueError("`warmup` should be an integer less than `iter`.")
    chains = int(chains)
    if chains < 1:
        raise ValueError("`chains` should be a positive integer.")

    iters = [iter] * chains
    thins = [thin] * chains
    warmups = [warmup] * chains

    # use chain_id argument if specified
    if kwargs.get('chain_id') is None:
        chain_id = list(range(chains))
    else:
        chain_id = [int(id) for id in kwargs['chain_id']]
        if len(set(chain_id)) != len(chain_id):
            raise ValueError("`chain_id` has duplicated elements.")
        chain_id_len = len(chain_id)
        if chain_id_len >= chains:
            chain_id = chain_id
        else:
            chain_id = chain_id + [max(chain_id) + 1 + i
                                   for i in range(chains - chain_id_len)]
        del kwargs['chain_id']

    inits_specified = False
    # slight difference here from rstan; Python's lists are not typed.
    if isinstance(init, Number):
        init = str(init)
    if isinstance(init, string_types):
        if init in ['0', 'random']:
            inits = [init] * chains
        else:
            inits = ["random"] * chains
        inits_specified = True
    if not inits_specified and isinstance(init, Callable):
        ## test if function takes argument named "chain_id"
        if "chain_id" in inspect.getargspec(init).args:
            inits = [init(chain_id=id) for id in chain_id]
        else:
            inits = [init()] * chains
        if not isinstance(inits[0], dict):
            raise ValueError("The function specifying initial values must "
                             "return a dictionary.")
        inits_specified = True
    if not inits_specified and isinstance(init, Sequence):
        if len(init) != chains:
            raise ValueError("Length of list of initial values does not "
                             "match number of chains.")
        if not all([isinstance(d, dict) for d in init]):
            raise ValueError("Initial value list is not a sequence of "
                             "dictionaries.")
        inits = init
        inits_specified = True
    if not inits_specified:
        raise ValueError("Invalid specification of initial values.")

    ## only one seed is needed by virtue of the RNG
    seed = _check_seed(seed)

    kwargs['method'] = "test_grad" if kwargs.get('test_grad') else 'sampling'

    all_control = {
        "adapt_engaged", "adapt_gamma", "adapt_delta", "adapt_kappa",
        "adapt_t0", "adapt_init_buffer", "adapt_term_buffer", "adapt_window",
        "stepsize", "stepsize_jitter", "metric", "int_time",
        "max_treedepth", "epsilon", "error", "inv_metric"
    }
    all_metrics = {"unit_e", "diag_e", "dense_e"}

    if control is not None:
        if not isinstance(control, dict):
            raise ValueError("`control` must be a dictionary")
        if not all(key in all_control for key in control):
            unknown = set(control) - all_control
            raise ValueError("`control` contains unknown parameters: {}".format(unknown))
        if control.get('metric') and control['metric'] not in all_metrics:
            raise ValueError("`metric` must be one of {}".format(all_metrics))
        kwargs['control'] = control

    argss = [dict() for _ in range(chains)]
    for i in range(chains):
        argss[i] = dict(chain_id=chain_id[i],
                        iter=iters[i], thin=thins[i], seed=seed,
                        warmup=warmups[i], init=inits[i],
                        algorithm=algorithm)

    if sample_file is not None:
        sample_file = _writable_sample_file(sample_file)
        if chains == 1:
            argss[0]['sample_file'] = sample_file
        elif chains > 1:
            for i in range(chains):
                argss[i]['sample_file'] = _append_id(sample_file, i)

    if diagnostic_file is not None:
        raise NotImplementedError("diagnostic_file not implemented yet.")

    if control is not None and "inv_metric" in control:
        inv_metric = control.pop("inv_metric")
        metric_dir = tempfile.mkdtemp()
        if isinstance(inv_metric, dict):
            for i in range(chains):
                if i not in inv_metric:
                    msg = "Invalid value for init_inv_metric found (keys={}). " \
                          "Use either a dictionary with chain_index as keys (0,1,2,...)" \
                          "or ndarray."
                    msg = msg.format(list(metric_file.keys()))
                    raise ValueError(msg)
                mass_values = inv_metric[i]
                metric_filename = "inv_metric_chain_{}.Rdata".format(str(i))
                metric_path = os.path.join(metric_dir, metric_filename)
                if isinstance(mass_values, str):
                    if not os.path.exists(mass_values):
                        raise ValueError("inverse metric file was not found: {}".format(mass_values))
                    shutil.copy(mass_values, metric_path)
                else:
                    stan_rdump(dict(inv_metric=mass_values), metric_path)
                argss[i]['metric_file'] = metric_path
        elif isinstance(inv_metric, str):
            if not os.path.exists(inv_metric):
                raise ValueError("inverse metric  file was not found: {}".format(inv_metric))
            for i in range(chains):
                metric_filename = "inv_metric_chain_{}.Rdata".format(str(i))
                metric_path = os.path.join(metric_dir, metric_filename)
                shutil.copy(inv_metric, metric_path)
                argss[i]['metric_file'] = metric_path
        elif isinstance(inv_metric, Iterable):
            metric_filename = "inv_metric_chain_0.Rdata"
            metric_path = os.path.join(metric_dir, metric_filename)
            stan_rdump(dict(inv_metric=inv_metric), metric_path)
            argss[0]['metric_file'] = metric_path
            for i in range(1, chains):
                metric_filename = "inv_metric_chain_{}.Rdata".format(str(i))
                metric_path = os.path.join(metric_dir, metric_filename)
                shutil.copy(argss[i-1]['metric_file'], metric_path)
                argss[i]['metric_file'] = metric_path
        else:
            argss[i]['metric_file'] = ""

    stepsize_list = None
    if "control" in kwargs and "stepsize" in kwargs["control"]:
        if isinstance(kwargs["control"]["stepsize"], Sequence):
            stepsize_list = kwargs["control"]["stepsize"]
            if len(kwargs["control"]["stepsize"]) == 1:
                kwargs["control"]["stepsize"] = kwargs["control"]["stepsize"][0]
            elif len(kwargs["control"]["stepsize"]) != chains:
                raise ValueError("stepsize length needs to equal chain count.")
            else:
                stepsize_list = kwargs["control"]["stepsize"]

    for i in range(chains):
        argss[i].update(kwargs)
        if stepsize_list is not None:
            argss[i]["control"]["stepsize"] = stepsize_list[i]
        argss[i] = _get_valid_stan_args(argss[i])

    return argss


def _get_valid_stan_args(base_args=None):
    """Fill in default values for arguments not provided in `base_args`.

    RStan does this in C++ in stan_args.hpp in the stan_args constructor.
    It seems easier to deal with here in Python.

    """
    args = base_args.copy() if base_args is not None else {}
    # Default arguments, c.f. rstan/rstan/inst/include/rstan/stan_args.hpp
    # values in args are going to be converted into C++ objects so
    # prepare them accordingly---e.g., unicode -> bytes -> std::string
    args['chain_id'] = args.get('chain_id', 1)
    args['append_samples'] = args.get('append_samples', False)
    if args.get('method') is None or args['method'] == "sampling":
        args['method'] = stan_args_method_t.SAMPLING
    elif args['method'] == "optim":
        args['method'] = stan_args_method_t.OPTIM
    elif args['method'] == 'test_grad':
        args['method'] = stan_args_method_t.TEST_GRADIENT
    elif args['method'] == 'variational':
        args['method'] = stan_args_method_t.VARIATIONAL
    else:
        args['method'] = stan_args_method_t.SAMPLING
    args['sample_file_flag'] = True if args.get('sample_file') else False
    args['sample_file'] = args.get('sample_file', '').encode('ascii')
    args['diagnostic_file_flag'] = True if args.get('diagnostic_file') else False
    args['diagnostic_file'] = args.get('diagnostic_file', '').encode('ascii')
    # NB: argument named "seed" not "random_seed"
    args['random_seed'] = args.get('seed', int(time.time()))

    args['metric_file_flag'] = True if args.get('metric_file') else False
    args['metric_file'] = args.get('metric_file', '').encode('ascii')

    if args['method'] == stan_args_method_t.VARIATIONAL:
        # variational does not use a `control` map like sampling
        args['ctrl'] = args.get('ctrl', dict(variational=dict()))
        args['ctrl']['variational']['iter'] = args.get('iter', 10000)
        args['ctrl']['variational']['grad_samples'] = args.get('grad_samples', 1)
        args['ctrl']['variational']['elbo_samples'] = args.get('elbo_samples', 100)
        args['ctrl']['variational']['eval_elbo'] = args.get('eval_elbo', 100)
        args['ctrl']['variational']['output_samples'] = args.get('output_samples', 1000)
        args['ctrl']['variational']['adapt_iter'] = args.get('adapt_iter', 50)
        args['ctrl']['variational']['eta'] = args.get('eta', 1.0)
        args['ctrl']['variational']['adapt_engaged'] = args.get('adapt_engaged', True)
        args['ctrl']['variational']['tol_rel_obj'] = args.get('tol_rel_obj', 0.01)
        if args.get('algorithm', '').lower() == 'fullrank':
            args['ctrl']['variational']['algorithm'] = variational_algo_t.FULLRANK
        else:
            args['ctrl']['variational']['algorithm'] = variational_algo_t.MEANFIELD
    elif args['method'] == stan_args_method_t.SAMPLING:
        args['ctrl'] = args.get('ctrl', dict(sampling=dict()))
        args['ctrl']['sampling']['iter'] = iter = args.get('iter', 2000)
        args['ctrl']['sampling']['warmup'] = warmup = args.get('warmup', iter // 2)
        calculated_thin = iter - warmup // 1000
        if calculated_thin < 1:
            calculated_thin = 1
        args['ctrl']['sampling']['thin'] = thin = args.get('thin', calculated_thin)
        args['ctrl']['sampling']['save_warmup'] = True  # always True now
        args['ctrl']['sampling']['iter_save_wo_warmup'] = iter_save_wo_warmup = 1 + (iter - warmup - 1) // thin
        args['ctrl']['sampling']['iter_save'] = iter_save_wo_warmup + 1 + (warmup - 1) // thin
        refresh = iter // 10 if iter >= 20 else 1
        args['ctrl']['sampling']['refresh'] = args.get('refresh', refresh)

        ctrl_lst = args.get('control', dict())
        ctrl_sampling = args['ctrl']['sampling']
        # NB: if these defaults change, remember to update docstrings
        ctrl_sampling['adapt_engaged'] = ctrl_lst.get("adapt_engaged", True)
        ctrl_sampling['adapt_gamma'] = ctrl_lst.get("adapt_gamma", 0.05)
        ctrl_sampling['adapt_delta'] = ctrl_lst.get("adapt_delta", 0.8)
        ctrl_sampling['adapt_kappa'] = ctrl_lst.get("adapt_kappa", 0.75)
        ctrl_sampling['adapt_t0'] = ctrl_lst.get("adapt_t0", 10.0)
        ctrl_sampling['adapt_init_buffer'] = ctrl_lst.get("adapt_init_buffer", 75)
        ctrl_sampling['adapt_term_buffer'] = ctrl_lst.get("adapt_term_buffer", 50)
        ctrl_sampling['adapt_window'] = ctrl_lst.get("adapt_window", 25)
        ctrl_sampling['stepsize'] = ctrl_lst.get("stepsize", 1.0)
        ctrl_sampling['stepsize_jitter'] = ctrl_lst.get("stepsize_jitter", 0.0)

        algorithm = args.get('algorithm', 'NUTS')
        if algorithm == 'HMC':
            args['ctrl']['sampling']['algorithm'] = sampling_algo_t.HMC
        elif algorithm == 'Metropolis':
            args['ctrl']['sampling']['algorithm'] = sampling_algo_t.Metropolis
        elif algorithm == 'NUTS':
            args['ctrl']['sampling']['algorithm'] = sampling_algo_t.NUTS
        elif algorithm == 'Fixed_param':
            args['ctrl']['sampling']['algorithm'] = sampling_algo_t.Fixed_param
            # TODO: Setting adapt_engaged to False solves the segfault reported
            # in issue #200; find out why this hack is needed. RStan deals with
            # the setting elsewhere.
            ctrl_sampling['adapt_engaged'] = False
        else:
            msg = "Invalid value for parameter algorithm (found {}; " \
                "require HMC, Metropolis, NUTS, or Fixed_param).".format(algorithm)
            raise ValueError(msg)

        metric = ctrl_lst.get('metric', 'diag_e')
        if metric == "unit_e":
            ctrl_sampling['metric'] = sampling_metric_t.UNIT_E
        elif metric == "diag_e":
            ctrl_sampling['metric'] = sampling_metric_t.DIAG_E
        elif metric == "dense_e":
            ctrl_sampling['metric'] = sampling_metric_t.DENSE_E

        if ctrl_sampling['algorithm'] == sampling_algo_t.NUTS:
            ctrl_sampling['max_treedepth'] = ctrl_lst.get("max_treedepth", 10)
        elif ctrl_sampling['algorithm'] == sampling_algo_t.HMC:
            ctrl_sampling['int_time'] = ctrl_lst.get('int_time', 6.283185307179586476925286766559005768e+00)
        elif ctrl_sampling['algorithm'] == sampling_algo_t.Metropolis:
            pass
        elif ctrl_sampling['algorithm'] == sampling_algo_t.Fixed_param:
            pass

    elif args['method'] == stan_args_method_t.OPTIM:
        args['ctrl'] = args.get('ctrl', dict(optim=dict()))
        args['ctrl']['optim']['iter'] = iter = args.get('iter', 2000)
        algorithm = args.get('algorithm', 'LBFGS')
        if algorithm == "BFGS":
            args['ctrl']['optim']['algorithm'] = optim_algo_t.BFGS
        elif algorithm == "Newton":
            args['ctrl']['optim']['algorithm'] = optim_algo_t.Newton
        elif algorithm == "LBFGS":
            args['ctrl']['optim']['algorithm'] = optim_algo_t.LBFGS
        else:
            msg = "Invalid value for parameter algorithm (found {}; " \
                  "require (L)BFGS or Newton).".format(algorithm)
            raise ValueError(msg)
        refresh = args['ctrl']['optim']['iter'] // 100
        args['ctrl']['optim']['refresh'] = args.get('refresh', refresh)
        if args['ctrl']['optim']['refresh'] < 1:
            args['ctrl']['optim']['refresh'] = 1
        args['ctrl']['optim']['init_alpha'] = args.get("init_alpha", 0.001)
        args['ctrl']['optim']['tol_obj'] = args.get("tol_obj", 1e-12)
        args['ctrl']['optim']['tol_grad'] = args.get("tol_grad", 1e-8)
        args['ctrl']['optim']['tol_param'] = args.get("tol_param", 1e-8)
        args['ctrl']['optim']['tol_rel_obj'] = args.get("tol_rel_obj", 1e4)
        args['ctrl']['optim']['tol_rel_grad'] = args.get("tol_rel_grad", 1e7)
        args['ctrl']['optim']['save_iterations'] = args.get("save_iterations", True)
        args['ctrl']['optim']['history_size'] = args.get("history_size", 5)
    elif args['method'] == stan_args_method_t.TEST_GRADIENT:
        args['ctrl'] = args.get('ctrl', dict(test_grad=dict()))
        args['ctrl']['test_grad']['epsilon'] = args.get("epsilon", 1e-6)
        args['ctrl']['test_grad']['error'] = args.get("error", 1e-6)

    init = args.get('init', "random")
    if isinstance(init, string_types):
        args['init'] = init.encode('ascii')
    elif isinstance(init, dict):
        args['init'] = "user".encode('ascii')
        # while the name is 'init_list', it is a dict; the name comes from rstan,
        # where list elements can have names
        args['init_list'] = init
    else:
        args['init'] = "random".encode('ascii')

    args['init_radius'] = args.get('init_r', 2.0)
    if (args['init_radius'] <= 0):
        args['init'] = b"0"

    # 0 initialization requires init_radius = 0
    if (args['init'] == b"0" or args['init'] == 0):
        args['init_radius'] = 0.0

    args['enable_random_init'] = args.get('enable_random_init', True)
    # RStan calls validate_args() here
    return args


def _check_seed(seed):
    """If possible, convert `seed` into a valid form for Stan (an integer
    between 0 and MAX_UINT, inclusive). If not possible, use a random seed
    instead and raise a warning if `seed` was not provided as `None`.
    """
    if isinstance(seed, (Number, string_types)):
        try:
            seed = int(seed)
        except ValueError:
            logger.warning("`seed` must be castable to an integer")
            seed = None
        else:
            if seed < 0:
                logger.warning("`seed` may not be negative")
                seed = None
            elif seed > MAX_UINT:
                raise ValueError('`seed` is too large; max is {}'.format(MAX_UINT))
    elif isinstance(seed, np.random.RandomState):
        seed = seed.randint(0, MAX_UINT)
    elif seed is not None:
        logger.warning('`seed` has unexpected type')
        seed = None

    if seed is None:
        seed = random.randint(0, MAX_UINT)

    return seed


def _organize_inits(inits, pars, dims):
    """Obtain a list of initial values for each chain.

    The parameter 'lp__' will be removed from the chains.

    Parameters
    ----------
    inits : list
        list of initial values for each chain.
    pars : list of str
    dims : list of list of int
        from (via cython conversion) vector[vector[uint]] dims

    Returns
    -------
    inits : list of dict

    """
    try:
        idx_of_lp = pars.index('lp__')
        del pars[idx_of_lp]
        del dims[idx_of_lp]
    except ValueError:
        pass
    starts = _calc_starts(dims)
    return [_par_vector2dict(init, pars, dims, starts) for init in inits]


def _calc_starts(dims):
    """Calculate starting indexes

    Parameters
    ----------
    dims : list of list of int
        from (via cython conversion) vector[vector[uint]] dims

    Examples
    --------
    >>> _calc_starts([[8, 2], [5], [6, 2]])
    [0, 16, 21]

    """
    # NB: Python uses 0-indexing; R uses 1-indexing.
    l = len(dims)
    s = [np.prod(d) for d in dims]
    starts = np.cumsum([0] + s)[0:l].tolist()
    # coerce things into ints before returning
    return [int(i) for i in starts]


def _par_vector2dict(v, pars, dims, starts=None):
    """Turn a vector of samples into an OrderedDict according to param dims.

    Parameters
    ----------
    y : list of int or float
    pars : list of str
        parameter names
    dims : list of list of int
        list of dimensions of parameters

    Returns
    -------
    d : dict

    Examples
    --------
    >>> v = list(range(31))
    >>> dims = [[5], [5, 5], []]
    >>> pars = ['mu', 'Phi', 'eta']
    >>> _par_vector2dict(v, pars, dims)  # doctest: +ELLIPSIS
    OrderedDict([('mu', array([0, 1, 2, 3, 4])), ('Phi', array([[ 5, ...

    """
    if starts is None:
        starts = _calc_starts(dims)
    d = OrderedDict()
    for i in range(len(pars)):
        l = int(np.prod(dims[i]))
        start = starts[i]
        end = start + l
        y = np.asarray(v[start:end])
        if len(dims[i]) > 1:
            y = y.reshape(dims[i], order='F')  # 'F' = Fortran, column-major
        d[pars[i]] = y.squeeze() if y.shape == (1,) else y
    return d


def _check_pars(allpars, pars):
    if len(pars) == 0:
        raise ValueError("No parameter specified (`pars` is empty).")
    for par in pars:
        if par not in allpars:
            raise ValueError("No parameter {}".format(par))


def _pars_total_indexes(names, dims, fnames, pars):
    """Obtain all the indexes for parameters `pars` in the sequence of names.

    `names` references variables that are in column-major order

    Parameters
    ----------
    names : sequence of str
        All the parameter names.
    dim : sequence of list of int
        Dimensions, in same order as `names`.
    fnames : sequence of str
        All the scalar parameter names
    pars : sequence of str
        The parameters of interest. It is assumed all elements in `pars` are in
        `names`.

    Returns
    -------
    indexes : OrderedDict of list of int
        Dictionary uses parameter names as keys. Indexes are column-major order.
        For each parameter there is also a key `par`+'_rowmajor' that stores the
        row-major indexing.

    Note
    ----
    Inside each parameter (vector or array), the sequence uses column-major
    ordering. For example, if we have parameters alpha and beta, having
    dimensions [2, 2] and [2, 3] respectively, the whole parameter sequence
    is alpha[0,0], alpha[1,0], alpha[0, 1], alpha[1, 1], beta[0, 0],
    beta[1, 0], beta[0, 1], beta[1, 1], beta[0, 2], beta[1, 2]. In short,
    like R matrix(..., bycol=TRUE).

    Example
    -------
    >>> pars_oi = ['mu', 'tau', 'eta', 'theta', 'lp__']
    >>> dims_oi = [[], [], [8], [8], []]
    >>> fnames_oi = ['mu', 'tau', 'eta[1]', 'eta[2]', 'eta[3]', 'eta[4]',
    ... 'eta[5]', 'eta[6]', 'eta[7]', 'eta[8]', 'theta[1]', 'theta[2]',
    ... 'theta[3]', 'theta[4]', 'theta[5]', 'theta[6]', 'theta[7]',
    ... 'theta[8]', 'lp__']
    >>> pars = ['mu', 'tau', 'eta', 'theta', 'lp__']
    >>> _pars_total_indexes(pars_oi, dims_oi, fnames_oi, pars)
    ... # doctest: +ELLIPSIS
    OrderedDict([('mu', (0,)), ('tau', (1,)), ('eta', (2, 3, ...

    """
    starts = _calc_starts(dims)

    def par_total_indexes(par):
        # if `par` is a scalar, it will match one of `fnames`
        if par in fnames:
            p = fnames.index(par)
            idx = tuple([p])
            return OrderedDict([(par, idx), (par+'_rowmajor', idx)])
        else:
            p = names.index(par)
            idx = starts[p] + np.arange(np.prod(dims[p]))
            idx_rowmajor = starts[p] + _idx_col2rowm(dims[p])
        return OrderedDict([(par, tuple(idx)), (par+'_rowmajor', tuple(idx_rowmajor))])

    indexes = OrderedDict()
    for par in pars:
        indexes.update(par_total_indexes(par))
    return indexes


def _idx_col2rowm(d):
    """Generate indexes to change from col-major to row-major ordering"""
    if 0 == len(d):
        return 1
    if 1 == len(d):
        return np.arange(d[0])
    # order='F' indicates column-major ordering
    idx = np.array(np.arange(np.prod(d))).reshape(d, order='F').T
    return idx.flatten(order='F')


def _get_kept_samples(n, sim):
    """Get samples to be kept from the chain(s) for `n`th parameter.

    Samples from different chains are merged.

    Parameters
    ----------
    n : int
    sim : dict
        A dictionary tied to a StanFit4Model instance.

    Returns
    -------
    samples : array
        Samples being kept, permuted and in column-major order.

    """
    return pystan._misc.get_kept_samples(n, sim)


def _get_samples(n, sim, inc_warmup=True):
    # NOTE: this is in stanfit-class.R in RStan (rather than misc.R)
    """Get chains for `n`th parameter.

    Parameters
    ----------
    n : int
    sim : dict
        A dictionary tied to a StanFit4Model instance.

    Returns
    -------
    chains : list of array
        Each chain is an element in the list.

    """
    return pystan._misc.get_samples(n, sim, inc_warmup)


def _redirect_stderr():
    """Redirect stderr for subprocesses to /dev/null

    Silences copious compilation messages.

    Returns
    -------
    orig_stderr : file descriptor
        Copy of original stderr file descriptor
    """
    sys.stderr.flush()
    stderr_fileno = sys.stderr.fileno()
    orig_stderr = os.dup(stderr_fileno)
    devnull = os.open(os.devnull, os.O_WRONLY)
    os.dup2(devnull, stderr_fileno)
    os.close(devnull)
    return orig_stderr


def _has_fileno(stream):
    """Returns whether the stream object seems to have a working fileno()

    Tells whether _redirect_stderr is likely to work.

    Parameters
    ----------
    stream : IO stream object

    Returns
    -------
    has_fileno : bool
        True if stream.fileno() exists and doesn't raise OSError or
        UnsupportedOperation
    """
    try:
        stream.fileno()
    except (AttributeError, OSError, IOError, io.UnsupportedOperation):
        return False
    return True


def _append_id(file, id, suffix='.csv'):
    fname = os.path.basename(file)
    fpath = os.path.dirname(file)
    fname2 = re.sub(r'\.csv\s*$', '_{}.csv'.format(id), fname)
    if fname2 == fname:
        fname2 = '{}_{}.csv'.format(fname, id)
    return os.path.join(fpath, fname2)


def _writable_sample_file(file, warn=True, wfun=None):
    """Check to see if file is writable, if not use temporary file"""
    if wfun is None:
        wfun = lambda x, y: '"{}" is not writable; use "{}" instead'.format(x, y)
    dir = os.path.dirname(file)
    dir = os.getcwd() if dir == '' else dir
    if os.access(dir, os.W_OK):
        return file
    else:
        dir2 = tempfile.mkdtemp()
        if warn:
            logger.warning(wfun(dir, dir2))
        return os.path.join(dir2, os.path.basename(file))


def is_legal_stan_vname(name):
    stan_kw1 = ('for', 'in', 'while', 'repeat', 'until', 'if', 'then', 'else',
                'true', 'false')
    stan_kw2 = ('int', 'real', 'vector', 'simplex', 'ordered', 'positive_ordered',
                'row_vector', 'matrix', 'corr_matrix', 'cov_matrix', 'lower', 'upper')
    stan_kw3 = ('model', 'data', 'parameters', 'quantities', 'transformed', 'generated')
    cpp_kw = ("alignas", "alignof", "and", "and_eq", "asm", "auto", "bitand", "bitor", "bool",
              "break", "case", "catch", "char", "char16_t", "char32_t", "class", "compl",
              "const", "constexpr", "const_cast", "continue", "decltype", "default", "delete",
              "do", "double", "dynamic_cast", "else", "enum", "explicit", "export", "extern",
              "false", "float", "for", "friend", "goto", "if", "inline", "int", "long", "mutable",
              "namespace", "new", "noexcept", "not", "not_eq", "nullptr", "operator", "or", "or_eq",
              "private", "protected", "public", "register", "reinterpret_cast", "return",
              "short", "signed", "sizeof", "static", "static_assert", "static_cast", "struct",
              "switch", "template", "this", "thread_local", "throw", "true", "try", "typedef",
              "typeid", "typename", "union", "unsigned", "using", "virtual", "void", "volatile",
              "wchar_t", "while", "xor", "xor_eq")
    illegal = stan_kw1 + stan_kw2 + stan_kw3 + cpp_kw
    if re.findall(r'(\.|^[0-9]|__$)', name):
        return False
    return not name in illegal


def _dict_to_rdump(data):
    parts = []
    for name, value in data.items():
        if isinstance(value, (Sequence, Number, np.number, np.ndarray, int, bool, float)) \
           and not isinstance(value, string_types):
            value = np.asarray(value)
        else:
            raise ValueError("Variable {} is not a number and cannot be dumped.".format(name))

        if value.dtype == np.bool:
            value = value.astype(int)

        if value.ndim == 0:
            s = '{} <- {}\n'.format(name, str(value))
        elif value.ndim == 1:
            s = '{} <-\nc({})\n'.format(name, ', '.join(str(v) for v in value))
        elif value.ndim > 1:
            tmpl = '{} <-\nstructure(c({}), .Dim = c({}))\n'
            # transpose value as R uses column-major
            # 'F' = Fortran, column-major
            s = tmpl.format(name,
                            ', '.join(str(v) for v in value.flatten(order='F')),
                            ', '.join(str(v) for v in value.shape))
        parts.append(s)
    return ''.join(parts)


def stan_rdump(data, filename):
    """
    Dump a dictionary with model data into a file using the R dump format that
    Stan supports.

    Parameters
    ----------
    data : dict
    filename : str

    """
    for name in data:
        if not is_legal_stan_vname(name):
            raise ValueError("Variable name {} is not allowed in Stan".format(name))
    with open(filename, 'w') as f:
        f.write(_dict_to_rdump(data))


def _rdump_value_to_numpy(s):
    """
    Convert a R dump formatted value to Numpy equivalent

    For example, "c(1, 2)" becomes ``array([1, 2])``

    Only supports a few R data structures. Will not work with European decimal format.
    """
    if "structure" in s:
        vector_str, shape_str = re.findall(r'c\([^\)]+\)', s)
        shape = [int(d) for d in shape_str[2:-1].split(',')]
        if '.' in vector_str:
            arr = np.array([float(v) for v in vector_str[2:-1].split(',')])
        else:
            arr = np.array([int(v) for v in vector_str[2:-1].split(',')])
        # 'F' = Fortran, column-major
        arr = arr.reshape(shape, order='F')
    elif "c(" in s:
        if '.' in s:
            arr = np.array([float(v) for v in s[2:-1].split(',')], order='F')
        else:
            arr = np.array([int(v) for v in s[2:-1].split(',')], order='F')
    else:
        arr = np.array(float(s) if '.' in s else int(s))
    return arr


def _remove_empty_pars(pars, pars_oi, dims_oi):
    """
    Remove parameters that are actually empty. For example, the parameter
    y would be removed with the following model code:

        transformed data { int n; n <- 0; }
        parameters { real y[n]; }

    Parameters
    ----------
    pars: iterable of str
    pars_oi: list of str
    dims_oi: list of list of int

    Returns
    -------
    pars_trimmed: list of str
    """
    pars = list(pars)
    for par, dim in zip(pars_oi, dims_oi):
        if par in pars and np.prod(dim) == 0:
            del pars[pars.index(par)]
    return pars


def read_rdump(filename):
    """
    Read data formatted using the R dump format

    Parameters
    ----------
    filename: str

    Returns
    -------
    data : OrderedDict
    """
    contents = open(filename).read().strip()
    names = [name.strip() for name in re.findall(r'^(\w+) <-', contents, re.MULTILINE)]
    values = [value.strip() for value in re.split('\w+ +<-', contents) if value]
    if len(values) != len(names):
        raise ValueError("Unable to read file. Unable to pair variable name with value.")
    d = OrderedDict()
    for name, value in zip(names, values):
        d[name.strip()] = _rdump_value_to_numpy(value.strip())
    return d

def to_dataframe(fit, pars=None, permuted=False, dtypes=None, inc_warmup=False, diagnostics=True, header=True):
    """Extract samples as a pandas dataframe for different parameters.

    Parameters
    ----------
    pars : {str, sequence of str}
        parameter (or quantile) name(s).
    permuted : bool
        If True, returned samples are permuted.
        If inc_warmup is True, warmup samples have negative order.
    dtypes : dict
        datatype of parameter(s).
        If nothing is passed, float will be used for all parameters.
    inc_warmup : bool
        If True, warmup samples are kept; otherwise they are
        discarded.
    diagnostics : bool
        If True, include hmc diagnostics in dataframe.
    header : bool
       If True, include header columns.

    Returns
    -------
    df : pandas dataframe
        Returned dataframe contains: [header_df]|[draws_df]|[diagnostics_df],
        where all groups are optional.
        To exclude draws_df use `pars=[]`.

    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("Pandas module not found. You can install pandas with: pip install pandas")

    fit._verify_has_samples()
    pars_original = pars
    if pars is None:
        pars = fit.sim['pars_oi']
    elif isinstance(pars, string_types):
        pars = [pars]
    if pars:
        pars = pystan.misc._remove_empty_pars(pars, fit.sim['pars_oi'], fit.sim['dims_oi'])
        allpars = fit.sim['pars_oi'] + fit.sim['fnames_oi']
        _check_pars(allpars, pars)

    if dtypes is None:
        dtypes = {}

    n_kept = [s if inc_warmup else s-w for s, w in zip(fit.sim['n_save'], fit.sim['warmup2'])]
    chains = len(fit.sim['samples'])

    diagnostic_type = {'divergent__':int,
                       'energy__':float,
                       'treedepth__':int,
                       'accept_stat__':float,
                       'stepsize__':float,
                       'n_leapfrog__':int}

    header_dict = OrderedDict()
    if header:
        idx = np.concatenate([np.full(n_kept[chain], chain, dtype=int) for chain in range(chains)])
        warmup = [np.zeros(n_kept[chain], dtype=np.int64) for chain in range(chains)]

        if inc_warmup:
            draw = []
            for chain, w in zip(range(chains), fit.sim['warmup2']):
                warmup[chain][:w] = 1
                draw.append(np.arange(n_kept[chain], dtype=np.int64) - w)
            draw = np.concatenate(draw)
        else:
            draw = np.concatenate([np.arange(n_kept[chain], dtype=np.int64) for chain in range(chains)])
        warmup = np.concatenate(warmup)

        header_dict = OrderedDict(zip(['chain', 'draw', 'warmup'], [idx, draw, warmup]))

    if permuted:
        if inc_warmup:
            chain_permutation = []
            chain_permutation_order = []
            permutation = []
            permutation_order = []
            for chain, p, w in zip(range(chains), fit.sim['permutation'], fit.sim['warmup2']):
                chain_permutation.append(list(range(-w, 0)) + p)
                chain_permutation_order.append(list(range(-w, 0)) + list(np.argsort(p)))
                permutation.append(sum(n_kept[:chain])+chain_permutation[-1]+w)
                permutation_order.append(sum(n_kept[:chain])+chain_permutation_order[-1]+w)
            chain_permutation = np.concatenate(chain_permutation)
            chain_permutation_order = np.concatenate(chain_permutation_order)
            permutation = np.concatenate(permutation)
            permutation_order = np.concatenate(permutation_order)

        else:
            chain_permutation = np.concatenate(fit.sim['permutation'])
            chain_permutation_order = np.concatenate([np.argsort(item) for item in fit.sim['permutation']])
            permutation = np.concatenate([sum(n_kept[:chain])+p for chain, p in enumerate(fit.sim['permutation'])])
            permutation_order = np.argsort(permutation)

        header_dict["permutation"] = permutation
        header_dict["chain_permutation"] = chain_permutation
        header_dict["permutation_order"] = permutation_order
        header_dict["chain_permutation_order"] = chain_permutation_order

    if header:
        header_df = pd.DataFrame.from_dict(header_dict)
    else:
        if permuted:
            header_df = pd.DataFrame.from_dict({"permutation_order" : header_dict["permutation_order"]})
        else:
            header_df = pd.DataFrame()

    fnames_set = set(fit.sim['fnames_oi'])
    pars_set = set(pars)
    if pars_original is None or fnames_set == pars_set:
        dfs = [pd.DataFrame.from_dict(pyholder.chains).iloc[-n:] for pyholder, n in zip(fit.sim['samples'], n_kept)]
        df = pd.concat(dfs, axis=0, sort=False, ignore_index=True)
        if dtypes:
            if not fnames_set.issuperset(pars_set):
                par_keys = OrderedDict([(par, []) for par in fit.sim['pars_oi']])
                for key in fit.sim['fnames_oi']:
                    par = key.split("[")
                    par = par[0]
                    par_keys[par].append(key)

            for par, dtype in dtypes.items():
                if isinstance(dtype, (float, np.float64)):
                    continue
                for key in par_keys.get(par, [par]):
                    df.loc[:, key] = df.loc[:, key].astype(dtype)

    elif pars:
        par_keys = dict()
        if not fnames_set.issuperset(pars_set):
            par_keys = OrderedDict([(par, []) for par in fit.sim['pars_oi']])
            for key in fit.sim['fnames_oi']:
                par = key.split("[")
                par = par[0]
                par_keys[par].append(key)

        columns = []
        for par in pars:
            columns.extend(par_keys.get(par, [par]))
        columns = list(np.unique(columns))

        df = pd.DataFrame(index=np.arange(sum(n_kept)), columns=columns, dtype=float)
        for key in columns:
            key_values = []
            for chain, (pyholder, n) in enumerate(zip(fit.sim['samples'], n_kept)):
                key_values.append(pyholder.chains[key][-n:])
            df.loc[:, key] = np.concatenate(key_values)

        for par, dtype in dtypes.items():
            if isinstance(dtype, (float, np.float64)):
                continue
            for key in par_keys.get(par, [par]):
                df.loc[:, key] = df.loc[:, key].astype(dtype)
    else:
        df = pd.DataFrame()

    if diagnostics:
        diagnostics_dfs = []
        for idx, (pyholder, permutation, n) in enumerate(zip(fit.sim['samples'], fit.sim['permutation'], n_kept), 1):
            diagnostics_df = pd.DataFrame(pyholder['sampler_params'], index=pyholder['sampler_param_names']).T
            diagnostics_df = diagnostics_df.iloc[-n:, :]
            for key, dtype in diagnostic_type.items():
                if key in diagnostics_df:
                    diagnostics_df.loc[:, key] = diagnostics_df.loc[:, key].astype(dtype)
            diagnostics_dfs.append(diagnostics_df)
        if diagnostics_dfs:
            diagnostics_df = pd.concat(diagnostics_dfs, axis=0, sort=False, ignore_index=True)
        else:
            diagnostics_df = pd.DataFrame()
    else:
        diagnostics_df = pd.DataFrame()

    df = pd.concat((header_df, df, diagnostics_df), axis=1, sort=False)
    if permuted:
        df.sort_values(by='permutation_order', inplace=True)
        if not header:
            df.drop(columns='permutation_order', inplace=True)
    return df

def get_stepsize(fit):
    """Parse stepsize from fit object

    Parameters
    ----------
    fit : StanFit4Model

    Returns
    -------
    list
        Returns an empty list if step sizes
        are not found in ``fit.get_adaptation_info``.
    """
    fit._verify_has_samples()
    stepsizes = []
    for adaptation_info in fit.get_adaptation_info():
        for line in adaptation_info.splitlines():
            if "Step size" in line:
                stepsizes.append(float(line.split("=")[1].strip()))
                break
    return stepsizes

def get_inv_metric(fit, as_dict=False):
    """Parse inverse metric from the fit object

    Parameters
    ----------
    fit : StanFit4Model
    as_dict : bool, optional

    Returns
    -------
    list or dict
        Returns an empty list if inverse metric
        is not found in ``fit.get_adaptation_info()``.
        If `as_dict` returns a dictionary which can be used with
        `.sampling` method.
    """
    fit._verify_has_samples()
    inv_metrics = []
    if not (("ctrl" in fit.stan_args[0]) and ("sampling" in fit.stan_args[0]["ctrl"])):
        return inv_metrics
    metric = [args["ctrl"]["sampling"]["metric"].name for args in fit.stan_args]
    for adaptation_info, metric_name in zip(fit.get_adaptation_info(), metric):
        iter_adaptation_info = iter(adaptation_info.splitlines())
        inv_metric_list = []
        for line in iter_adaptation_info:
            if any(value in line for value in ["Step size", "Adaptation"]):
                continue
            elif "inverse mass matrix" in line:
                for line in iter_adaptation_info:
                    stripped_set = set(line.replace("# ", "").replace(" ", "").replace(",", ""))
                    if stripped_set.issubset(set(".-1234567890e")):
                        inv_metric = np.array(list(map(float, line.replace("# ", "").strip().split(","))))
                        if metric_name == "DENSE_E":
                            inv_metric = np.atleast_2d(inv_metric)
                        inv_metric_list.append(inv_metric)
                    else:
                        break
        inv_metrics.append(np.concatenate(inv_metric_list))
    return inv_metrics if not as_dict else dict(enumerate(inv_metrics))

def get_last_position(fit, warmup=False):
    """Parse last position from fit object

    Parameters
    ----------
    fit : StanFit4Model
    warmup : bool
        If True, returns the last warmup position, when warmup has been done.
        Otherwise function returns the first sample position.

    Returns
    -------
    list
        list contains a dictionary of last draw from each chain.
    """
    fit._verify_has_samples()
    positions = []
    extracted = fit.extract(permuted=False, pars=fit.model_pars, inc_warmup=warmup)

    draw_location = -1
    if warmup:
        draw_location += max(1, fit.sim["warmup"])

    chains = fit.sim["chains"]
    for i in range(chains):
        extract_pos = {key : values[draw_location, i] for key, values in extracted.items()}
        positions.append(extract_pos)
    return positions



MAX_UINT = 2**31 - 1  # conservative choice, maximum unsigned int on 32-bit Python 2.7

try:
    from enum import Enum  # Python 3.4
except ImportError:
    from pystan.external.enum import Enum

sampling_algo_t = Enum('sampling_algo_t', 'NUTS HMC Metropolis Fixed_param', module=__name__)
variational_algo_t = Enum('variational_algo_t', 'MEANFIELD FULLRANK', module=__name__)


class optim_algo_t(Enum):
    Newton = 1
    BFGS = 3
    LBFGS = 4

sampling_metric_t = Enum('sampling_metric_t', 'UNIT_E DIAG_E DENSE_E', module=__name__)
stan_args_method_t = Enum('stan_args_method_t', 'SAMPLING OPTIM TEST_GRADIENT VARIATIONAL', module=__name__)



# distutils: language = c++
# cython: language_level=2
#-----------------------------------------------------------------------------
# Copyright (c) 2013, Allen B. Riddell
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------
from libcpp cimport bool
from pystan.stanc cimport PyStancResult, stanc as c_stanc
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

def stanc(bytes model_stancode, bytes model_name,
          bool allow_undefined, bytes filename,
          vector[string] include_paths):
    cdef PyStancResult result
    c_stanc(model_stancode, model_name, allow_undefined,
            filename, include_paths, result)
    result_include_paths = []
    for include_path in result.include_paths:
        result_include_paths.append(include_path.decode('utf-8'))
    return {'status': result.status,
            'msg': result.msg.decode('utf-8'),
            'model_cppname': result.model_cppname.decode('ascii'),
            'cppcode': result.cppcode.decode('ascii'),
            'include_paths' : result_include_paths,
            }



# distutils: language = c++
#-----------------------------------------------------------------------------
# Copyright (c) 2013, Allen B. Riddell
#
# This file is licensed under Version 3.0 of the GNU General Public
# License. See LICENSE for a text of the license.
#-----------------------------------------------------------------------------

from libcpp cimport bool
from libcpp.map cimport map
from libcpp.pair cimport pair
from libcpp.string cimport string
from libcpp.vector cimport vector

ctypedef unsigned int uint  # needed for templates

cdef extern from "stan_fit.hpp" namespace "pystan":
    ctypedef map[string, pair[vector[double], vector[size_t]]] vars_r_t
    ctypedef map[string, pair[vector[int], vector[size_t]]] vars_i_t
    ctypedef enum sampling_algo_t:
        NUTS = 1
        HMC = 2
        Metropolis = 3
        Fixed_param = 4
    ctypedef enum optim_algo_t:
        Newton = 1
        BFGS = 3
        LBFGS = 4
    ctypedef enum variational_algo_t:
        MEANFIELD = 1
        FULLRANK = 2
    ctypedef enum sampling_metric_t:
        UNIT_E = 1
        DIAG_E = 2
        DENSE_E = 3
    ctypedef enum stan_args_method_t:
        SAMPLING = 1
        OPTIM = 2
        TEST_GRADIENT = 3
        VARIATIONAL = 4

    cdef cppclass stan_fit[M, R]:
        stan_fit(vars_r_t& vars_r, vars_i_t& vars_i, uint random_seed) except +
        bool update_param_oi(vector[string] pars)
        vector[double] unconstrain_pars(vars_r_t& vars_r, vars_i_t& vars_i) except +
        vector[double] constrain_pars(vector[double]& params_r) except +
        double log_prob(vector[double] upar, bool jacobian_adjust_transform, bool gradient) except +
        vector[double] grad_log_prob(vector[double] upar, bool jacobian_adjust_transform) except +
        int num_pars_unconstrained()
        vector[string] unconstrained_param_names(bool include_tparams, bool include_gqs)
        vector[string] constrained_param_names(bool include_tparams, bool include_gqs)
        int call_sampler(StanArgs&, StanHolder&) except +
        vector[string] param_names()
        vector[string] param_names_oi()
        vector[vector[uint]] param_dims()
        vector[vector[uint]] param_dims_oi()
        vector[string] param_fnames_oi()

    cdef struct sampling_t:
        int iter
        int refresh
        sampling_algo_t algorithm
        int warmup
        int thin
        bool save_warmup  # whether to save warmup samples (always true now)
        int iter_save # number of iterations saved
        int iter_save_wo_warmup  # number of iterations saved wo warmup
        bool adapt_engaged
        double adapt_gamma
        double adapt_delta
        double adapt_kappa
        uint adapt_init_buffer
        uint adapt_term_buffer
        uint adapt_window
        double adapt_t0
        sampling_metric_t metric  # UNIT_E, DIAG_E, DENSE_E
        # string init_inv_metric
        double stepsize  # default to 1
        double stepsize_jitter
        int max_treedepth  # for NUTS, default to 10.
        double int_time  # for HMC, default to 2 * pi

    cdef struct optim_t:
        int iter  # default to 2000
        int refresh  # default to 100
        optim_algo_t algorithm  # Newton (L)BFGS
        bool save_iterations  # default to false
        double init_alpha
        double tol_obj
        double tol_grad
        double tol_param
        double tol_rel_obj
        double tol_rel_grad
        int history_size

    cdef struct variational_t:
        int iter
        variational_algo_t algorithm
        int grad_samples
        int elbo_samples
        int eval_elbo
        int output_samples
        double eta
        bool adapt_engaged
        int adapt_iter
        double tol_rel_obj

    cdef struct test_grad_t:
        double epsilon # default to 1e-6, for test_grad
        double error # default to 1e-6, for test_grad

    cdef union ctrl_t:
        sampling_t sampling
        optim_t optim
        variational_t variational
        test_grad_t test_grad

    cdef cppclass StanArgs:
        uint random_seed
        uint chain_id
        string init
        map[string, pair[vector[double], vector[size_t] ] ] init_vars_r
        map[string, pair[vector[int], vector[size_t] ] ] init_vars_i
        double init_radius
        bool enable_random_init  # enable randomly partially specifying inits
        string sample_file
        bool append_samples
        bool sample_file_flag
        stan_args_method_t method
        string diagnostic_file
        bool diagnostic_file_flag
        ctrl_t ctrl
        string metric_file
        bool metric_file_flag

    cdef cppclass StanHolder:
        int num_failed
        bool test_grad
        vector[double] inits
        vector[double] par
        double value
        vector[vector[double] ] chains
        vector[string] chain_names
        StanArgs args
        vector[double] mean_pars
        vector[string] mean_par_names
        double mean_lp__
        string adaptation_info
        vector[vector[double] ] sampler_params
        vector[string] sampler_param_names

    # free standing functions
    void get_all_flatnames(vector[string] names, vector[vector[uint]]& dims, vector[string] fnames, bool col_major)
